"""Helpers for running code snippets from the XGBoost documentation."""

import atexit
import json
import subprocess
import tempfile
import uuid
from pathlib import Path

from docutils import nodes
from sphinx.application import Sphinx

PROJECT_ROOT = Path(__file__).resolve().parent.parent
NO_DOCTEST_CLASS = "no-doctest"
_R_SESSIONS: dict[str, "RSession"] = {}


class RSession:
    """A persistent R process for doctest snippets from one document."""

    def __init__(self) -> None:
        # pylint: disable=consider-using-with
        self._tmpdir = tempfile.TemporaryDirectory()
        self._proc = subprocess.Popen(
            ["R", "--vanilla", "--slave"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=PROJECT_ROOT,
            text=True,
            bufsize=1,
        )

    def run(self, code: str) -> None:
        """Run a snippet of code."""
        if self._proc.poll() is not None:
            raise RuntimeError(f"R process exited with code {self._proc.returncode}.")
        assert self._proc.stdin is not None

        path = Path(self._tmpdir.name) / "xgboost-doc-test.R"
        path.write_text(code, encoding="utf-8")
        # The marker is printed after the sourced snippet with its status code.  Python
        # reads R stdout until this marker, so each snippet can be checked without
        # restarting the R process.
        marker = f"__XGBOOST_R_DOCTEST_DONE_{uuid.uuid4().hex}__"
        command = rf""".__xgb_doctest_status <- 0L
tryCatch(
  source({json.dumps(str(path))}, local = globalenv(), echo = FALSE, keep.source = FALSE),
  error = function(e) {{
    .__xgb_doctest_status <<- 1L
    cat(conditionMessage(e), '\n', file = stderr())
  }}
)
cat('\n{marker}:', .__xgb_doctest_status, '\n', sep = '')
rm(.__xgb_doctest_status)
flush.console()
"""
        self._proc.stdin.write(command)
        self._proc.stdin.flush()

        output, failed = self._read_until_marker(marker)
        if not failed:
            return

        msg = ["R documentation snippet failed."]
        if output:
            msg.extend(["output:", output])
        raise RuntimeError("\n".join(msg))

    def close(self) -> None:
        """Close the R session."""
        if self._proc.poll() is None:
            # quit if R is still running.
            assert self._proc.stdin is not None
            self._proc.stdin.write("q('no', status = 0, runLast = FALSE)\n")
            self._proc.stdin.flush()
            self._proc.wait()
        self._tmpdir.cleanup()

    def _read_until_marker(self, marker: str) -> tuple[str, bool]:
        assert self._proc.stdout is not None
        output: list[str] = []
        while True:
            line = self._proc.stdout.readline()
            if line == "":
                raise RuntimeError(
                    "R process exited before reporting snippet status. "
                    f"Exit code: {self._proc.poll()}."
                )
            if line.startswith(f"{marker}:"):
                return "".join(output), line.removeprefix(f"{marker}:").strip() != "0"
            output.append(line)


def run_r_code(session_key: str, code: str) -> None:
    """Run one R doctest snippet in a persistent document session."""
    session = _R_SESSIONS.get(session_key)
    if session is None:
        session = RSession()
        _R_SESSIONS[session_key] = session
    session.run(code)


def close_r_session(session_key: str) -> None:
    """Close the R session for a document if it was started."""
    session = _R_SESSIONS.pop(session_key, None)
    if session is not None:
        session.close()


def close_all_r_sessions() -> None:
    """Close all R doctest sessions."""
    for session_key in list(_R_SESSIONS):
        close_r_session(session_key)


def _normalise_language(language: str | None) -> str | None:
    if language is None:
        return None
    language = language.lower()
    if language in {"py", "python", "python3"}:
        return "python"
    if language in {"r", "s", "rlang"}:
        return "r"
    return None


def _mark_python_node(node: nodes.literal_block, docname: str) -> None:
    node["testnodetype"] = "testcode"
    node["groups"] = [docname]
    node["options"] = {}


def _mark_r_node(
    node: nodes.literal_block, docname: str, *, close_after: bool = False
) -> None:
    code = node.astext()
    test = [
        "from xgboost_doc_doctest import run_r_code as _xgb_run_r_code",
        f"_xgb_run_r_code({docname!r}, {code!r})",
    ]
    if close_after:
        test.extend(
            [
                "from xgboost_doc_doctest import close_r_session as _xgb_close_r_session",
                f"_xgb_close_r_session({docname!r})",
            ]
        )
    node["test"] = "\n".join(test)
    node["testnodetype"] = "testcode"
    node["groups"] = [docname]
    node["options"] = {}


def _has_class(node: nodes.Node, class_name: str) -> bool:
    return class_name in node.get("classes", [])


def _has_ancestor_class(node: nodes.Node, class_name: str) -> bool:
    parent = node.parent
    while parent is not None:
        if _has_class(parent, class_name):
            return True
        parent = parent.parent
    return False


def mark_doctest_nodes(app: Sphinx, doctree: nodes.document) -> None:
    """Mark code-tab blocks so ``sphinx.ext.doctest`` executes them."""
    docname = app.env.docname
    r_nodes: list[nodes.literal_block] = []
    for node in doctree.findall(nodes.literal_block):
        if "testnodetype" in node:
            continue
        if _has_class(node, NO_DOCTEST_CLASS) or _has_ancestor_class(
            node, NO_DOCTEST_CLASS
        ):
            continue
        if not _has_ancestor_class(node, "code-tab"):
            continue

        language = _normalise_language(node.get("language"))
        if language == "python":
            _mark_python_node(node, docname)
        elif language == "r":
            r_nodes.append(node)

    for node in r_nodes[:-1]:
        _mark_r_node(node, docname)
    if r_nodes:
        _mark_r_node(r_nodes[-1], docname, close_after=True)


def setup(app: Sphinx) -> dict[str, bool]:
    """Entry point for sphinx app setup."""
    app.connect("doctree-read", mark_doctest_nodes)
    atexit.register(close_all_r_sessions)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Testing code shared by other tests."""

import collections
import importlib.util
import json
import os
import tempfile
from typing import Any, Callable, Dict, Type

import numpy as np

import xgboost as xgb
from xgboost._typing import ArrayLike


def validate_leaf_output(leaf: np.ndarray, num_parallel_tree: int) -> None:
    """Validate output for predict leaf tests."""
    for i in range(leaf.shape[0]):  # n_samples
        for j in range(leaf.shape[1]):  # n_rounds
            for k in range(leaf.shape[2]):  # n_classes
                tree_group = leaf[i, j, k, :]
                assert tree_group.shape[0] == num_parallel_tree
                # No sampling, all trees within forest are the same
                assert np.all(tree_group == tree_group[0])


def validate_data_initialization(
    dmatrix: Type, model: Type[xgb.XGBModel], X: ArrayLike, y: ArrayLike
) -> None:
    """Assert that we don't create duplicated DMatrix."""

    old_init = dmatrix.__init__
    count = [0]

    def new_init(self: Any, **kwargs: Any) -> Callable:
        count[0] += 1
        return old_init(self, **kwargs)

    dmatrix.__init__ = new_init
    model(n_estimators=1).fit(X, y, eval_set=[(X, y)])

    assert count[0] == 1
    count[0] = 0  # only 1 DMatrix is created.

    y_copy = y.copy()
    model(n_estimators=1).fit(X, y, eval_set=[(X, y_copy)])
    assert count[0] == 2  # a different Python object is considered different

    dmatrix.__init__ = old_init


# pylint: disable=too-many-arguments,too-many-locals
def get_feature_weights(
    *,
    X: ArrayLike,
    y: ArrayLike,
    fw: np.ndarray,
    parser_path: str,
    tree_method: str,
    model: Type[xgb.XGBModel] = xgb.XGBRegressor,
) -> np.ndarray:
    """Get feature weights using the demo parser."""
    with tempfile.TemporaryDirectory() as tmpdir:
        colsample_bynode = 0.5
        reg = model(
            tree_method=tree_method,
            colsample_bynode=colsample_bynode,
            feature_weights=fw,
        )

        reg.fit(X, y)
        model_path = os.path.join(tmpdir, "model.json")
        reg.save_model(model_path)
        with open(model_path, "r", encoding="utf-8") as fd:
            model = json.load(fd)

        spec = importlib.util.spec_from_file_location("JsonParser", parser_path)
        assert spec is not None
        jsonm = importlib.util.module_from_spec(spec)
        assert spec.loader is not None
        spec.loader.exec_module(jsonm)
        model = jsonm.Model(model)
        splits: Dict[int, int] = {}
        total_nodes = 0
        for tree in model.trees:
            n_nodes = len(tree.nodes)
            total_nodes += n_nodes
            for n in range(n_nodes):
                if tree.is_leaf(n):
                    continue
                if splits.get(tree.split_index(n), None) is None:
                    splits[tree.split_index(n)] = 1
                else:
                    splits[tree.split_index(n)] += 1

        od = collections.OrderedDict(sorted(splits.items()))
        tuples = list(od.items())
        k, v = list(zip(*tuples))
        w = np.polyfit(k, v, deg=1)
        return w



"""
Wrapper script to run a command inside a Docker container
"""

import argparse
import grp
import itertools
import os
import pathlib
import pwd
import subprocess
import sys
import textwrap

OPS_DIR = pathlib.Path(__file__).expanduser().resolve().parent
PROJECT_ROOT_DIR = OPS_DIR.parent
LINEWIDTH = 88
TEXT_WRAPPER = textwrap.TextWrapper(
    width=LINEWIDTH,
    initial_indent="",
    subsequent_indent="    ",
    break_long_words=False,
    break_on_hyphens=False,
)


def parse_run_args(*, raw_run_args: str) -> list[str]:
    return [x for x in raw_run_args.split() if x]


def get_user_ids() -> dict[str, str]:
    uid = os.getuid()
    gid = os.getgid()
    return {
        "CI_BUILD_UID": str(uid),
        "CI_BUILD_USER": pwd.getpwuid(uid).pw_name,
        "CI_BUILD_GID": str(gid),
        "CI_BUILD_GROUP": grp.getgrgid(gid).gr_name,
    }


def fancy_print_cli_args(*, cli_args: list[str]) -> None:
    print(
        "=" * LINEWIDTH
        + "\n"
        + "  \\\n".join(TEXT_WRAPPER.wrap(" ".join(cli_args)))
        + "\n"
        + "=" * LINEWIDTH
        + "\n",
        flush=True,
    )


def docker_run(
    *,
    image_uri: str,
    command_args: list[str],
    use_gpus: bool,
    workdir: pathlib.Path,
    user_ids: dict[str, str],
    extra_args: list[str],
) -> None:
    # Command-line arguments to be passed to `docker run`
    docker_run_cli_args = ["--rm", "--pid=host"]

    if use_gpus:
        docker_run_cli_args.extend(["--gpus", "all"])

    docker_run_cli_args.extend(["-v", f"{workdir}:/workspace", "-w", "/workspace"])
    docker_run_cli_args.extend(
        itertools.chain.from_iterable([["-e", f"{k}={v}"] for k, v in user_ids.items()])
    )
    docker_run_cli_args.extend(extra_args)
    docker_run_cli_args.append(image_uri)
    docker_run_cli_args.extend(command_args)

    cli_args = ["docker", "run"] + docker_run_cli_args
    fancy_print_cli_args(cli_args=cli_args)
    subprocess.run(cli_args, check=True, encoding="utf-8")


def main(*, args: argparse.Namespace) -> None:
    run_args = parse_run_args(raw_run_args=args.run_args)
    user_ids = get_user_ids()

    if args.use_gpus:
        print("Using NVIDIA GPUs for `docker run`")
    if args.interactive:
        print("Using interactive mode for `docker run`")
        run_args.append("-it")

    docker_run(
        image_uri=args.image_uri,
        command_args=args.command_args,
        use_gpus=args.use_gpus,
        workdir=args.workdir,
        user_ids=user_ids,
        extra_args=run_args,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        usage=(
            f"{sys.argv[0]} --image-uri IMAGE_URI [--use-gpus] [--interactive] "
            "[--workdir WORKDIR] [--run-args RUN_ARGS] -- COMMAND_ARG "
            "[COMMAND_ARG ...]"
        ),
        description="Run tasks inside a Docker container",
    )
    parser.add_argument(
        "--image-uri",
        type=str,
        required=True,
        help=(
            "Fully qualified image URI to identify the container, e.g. "
            "492475357299.dkr.ecr.us-west-2.amazonaws.com/xgb-ci.gpu:main"
        ),
    )
    parser.add_argument(
        "--use-gpus",
        action="store_true",
        help=(
            "Grant the container access to NVIDIA GPUs; requires the NVIDIA "
            "Container Toolkit."
        ),
    )
    parser.add_argument(
        "--interactive",
        action="store_true",
        help=(
            "Run the container in the interactive mode; requires an interactive shell "
            "(TTY). With this flag, you can use Ctrl-C to interrupt an long-running "
            "command."
        ),
    )
    parser.add_argument(
        "--workdir",
        type=lambda p: pathlib.Path(p).expanduser().resolve(),
        default=PROJECT_ROOT_DIR,
        help="Path to working directory; if unset, use the project's root",
    )
    parser.add_argument(
        "--run-args",
        type=str,
        default="",
        help=(
            "Argument(s) to be passed to `docker run`. When passing multiple "
            "arguments, use single quotes to wrap them. Example: "
            "--run-args '--cap-add SYS_PTRACE --shm-size=4g'"
        ),
    )
    parser.add_argument(
        "command_args",
        metavar="COMMAND_ARG",
        type=str,
        nargs="+",
        help=(
            "Argument(s) for the command to execute. NOTE. Make sure to specify "
            "double-dash (--) to clearly distinguish between the command and the "
            "preceding parameters. Example: --run-args '--cap-add SYS_PTRACE "
            "--shm-size=4g' -- ./myprog"
        ),
    )

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    parsed_args = parser.parse_args()
    main(args=parsed_args)



"""XGBoost collective communication related API."""

import ctypes
import logging
import os
import pickle
from dataclasses import dataclass
from enum import IntEnum, unique
from typing import Any, Callable, Dict, Optional, TypeAlias, Union

import numpy as np

from ._typing import _T
from .core import _LIB, _check_call, build_info, c_str, make_jcargs, py_str

LOGGER = logging.getLogger("[xgboost.collective]")


_Conf: TypeAlias = Dict[str, Union[int, str]]
_ArgVals: TypeAlias = Optional[Union[int, str]]
_Args: TypeAlias = Dict[str, _ArgVals]


@dataclass
class Config:
    """User configuration for the communicator context. This is used for easier
    integration with distributed frameworks. Users of the collective module can pass the
    parameters directly into tracker and the communicator.

    .. versionadded:: 3.0

    Attributes
    ----------
    retry : See `dmlc_retry` in :py:meth:`init`.

    timeout :
        See `dmlc_timeout` in :py:meth:`init`. This is only used for communicators, not
        the tracker. They are different parameters since the timeout for tracker limits
        only the time for starting and finalizing the communication group, whereas the
        timeout for communicators limits the time used for collective operations, like
        :py:meth:`allreduce`.

    tracker_host_ip : See :py:class:`~xgboost.tracker.RabitTracker`.

    tracker_port : See :py:class:`~xgboost.tracker.RabitTracker`.

    tracker_timeout : See :py:class:`~xgboost.tracker.RabitTracker`.

    worker_port :

        The port each worker listens to for peer-to-peer connections. By default,
        workers use an available port assigned by the OS. This option can be used in
        restricted network environments where only specific ports are open.

        This can be an integer for a fixed port used by all workers, or a callback
        function that takes no arguments and returns a port number. The callback is
        invoked per-worker at the worker side.

        .. note::

            The option does not affect the NCCL communicator group, which must be
            configured via NCCL's own environment variables.

    """

    retry: Optional[int] = None
    timeout: Optional[int] = None

    tracker_host_ip: Optional[str] = None
    tracker_port: Optional[int] = None
    tracker_timeout: Optional[int] = None

    worker_port: Optional[Union[Callable[[], int], int]] = None

    def update_worker_args(self, args: _Conf) -> _Conf:
        """Worker side arguments resolution."""
        if self.worker_port is None:
            return args
        if callable(self.worker_port):
            args["dmlc_worker_port"] = self.worker_port()
        else:
            args["dmlc_worker_port"] = self.worker_port
        return args

    def get_comm_config(self, args: _Conf) -> _Conf:
        """Update the arguments for the communicator."""
        if self.retry is not None:
            args["dmlc_retry"] = self.retry
        if self.timeout is not None:
            args["dmlc_timeout"] = self.timeout
        return args


def init(**args: _ArgVals) -> None:
    """Initialize the collective library with arguments.

    Parameters
    ----------
    args :
        Keyword arguments representing the parameters and their values.

        Accepted parameters:
          - dmlc_communicator: The type of the communicator.
            * rabit: Use Rabit. This is the default if the type is unspecified.
            * federated: Use the gRPC interface for Federated Learning.

        Only applicable to the Rabit communicator:
          - dmlc_tracker_uri: Hostname of the tracker.
          - dmlc_tracker_port: Port number of the tracker.
          - dmlc_task_id: ID of the current task, can be used to obtain deterministic
          - dmlc_retry: The number of retry when handling network errors.
          - dmlc_timeout: Timeout in seconds.
          - dmlc_nccl_path: Path to load (dlopen) nccl for GPU-based communication.

        Only applicable to the Federated communicator:
          - federated_server_address: Address of the federated server.
          - federated_world_size: Number of federated workers.
          - federated_rank: Rank of the current worker.
          - federated_server_cert: Server certificate file path. Only needed for the SSL
            mode.
          - federated_client_key: Client key file path. Only needed for the SSL mode.
          - federated_client_cert: Client certificate file path. Only needed for the SSL
            mode.

        Use upper case for environment variables, use lower case for runtime
        configuration.

    """
    _check_call(_LIB.XGCommunicatorInit(make_jcargs(**args)))


def finalize() -> None:
    """Finalize the communicator."""
    _check_call(_LIB.XGCommunicatorFinalize())


def get_rank() -> int:
    """Get rank of current process.

    Returns
    -------
    rank : int
        Rank of current process.
    """
    ret = _LIB.XGCommunicatorGetRank()
    return ret


def get_world_size() -> int:
    """Get total number workers.

    Returns
    -------
    n :
        Total number of process.
    """
    ret = _LIB.XGCommunicatorGetWorldSize()
    return ret


def is_distributed() -> bool:
    """If the collective communicator is distributed."""
    is_dist = _LIB.XGCommunicatorIsDistributed()
    return bool(is_dist)


def communicator_print(msg: Any) -> None:
    """Print message to the communicator.

    This function can be used to communicate the information of
    the progress to the communicator.

    Parameters
    ----------
    msg : str
        The message to be printed to the communicator.
    """
    if not isinstance(msg, str):
        msg = str(msg)
    is_dist = _LIB.XGCommunicatorIsDistributed()
    if is_dist != 0:
        _check_call(_LIB.XGCommunicatorPrint(c_str(msg.strip())))
    else:
        print(msg.strip(), flush=True)


def get_processor_name() -> str:
    """Get the processor name.

    Returns
    -------
    name :
        The name of processor(host)
    """
    name_str = ctypes.c_char_p()
    _check_call(_LIB.XGCommunicatorGetProcessorName(ctypes.byref(name_str)))
    value = name_str.value
    return py_str(value)


def broadcast(data: _T, root: int) -> _T:
    """Broadcast object from one node to all other nodes.

    Parameters
    ----------
    data : any type that can be pickled
        Input data, if current rank does not equal root, this can be None
    root : int
        Rank of the node to broadcast data from.

    Returns
    -------
    object : int
        the result of broadcast.
    """
    rank = get_rank()
    length = ctypes.c_ulong()
    if root == rank:
        assert data is not None, "need to pass in data when broadcasting"
        s = pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL)
        length.value = len(s)
    # Run first broadcast
    _check_call(
        _LIB.XGCommunicatorBroadcast(
            ctypes.byref(length), ctypes.sizeof(ctypes.c_ulong), root
        )
    )
    if root != rank:
        dptr = (ctypes.c_char * length.value)()
        # run second
        _check_call(
            _LIB.XGCommunicatorBroadcast(
                ctypes.cast(dptr, ctypes.c_void_p), length.value, root
            )
        )
        data = pickle.loads(dptr.raw)
        del dptr
    else:
        _check_call(
            _LIB.XGCommunicatorBroadcast(
                ctypes.cast(ctypes.c_char_p(s), ctypes.c_void_p), length.value, root
            )
        )
        del s
    return data


# enumeration of dtypes
def _map_dtype(dtype: np.dtype) -> int:
    dtype_map = {
        np.dtype("float16"): 0,
        np.dtype("float32"): 1,
        np.dtype("float64"): 2,
        np.dtype("int8"): 4,
        np.dtype("int16"): 5,
        np.dtype("int32"): 6,
        np.dtype("int64"): 7,
        np.dtype("uint8"): 8,
        np.dtype("uint16"): 9,
        np.dtype("uint32"): 10,
        np.dtype("uint64"): 11,
    }
    try:
        dtype_map.update({np.dtype("float128"): 3})
    except TypeError:  # float128 doesn't exist on the system
        pass

    if dtype not in dtype_map:
        raise TypeError(f"data type {dtype} is not supported on the current platform.")

    return dtype_map[dtype]


@unique
class Op(IntEnum):
    """Supported operations for allreduce."""

    MAX = 0
    MIN = 1
    SUM = 2
    BITWISE_AND = 3
    BITWISE_OR = 4
    BITWISE_XOR = 5


def allreduce(data: np.ndarray, op: Op) -> np.ndarray:
    """Perform allreduce, return the result.

    Parameters
    ----------
    data :
        Input data.
    op :
        Reduction operator.

    Returns
    -------
    result :
        The result of allreduce, have same shape as data

    Notes
    -----
    This function is not thread-safe.
    """
    if not isinstance(data, np.ndarray):
        raise TypeError("allreduce only takes in numpy.ndarray")
    buf = data.ravel().copy()
    _check_call(
        _LIB.XGCommunicatorAllreduce(
            buf.ctypes.data_as(ctypes.c_void_p),
            buf.size,
            _map_dtype(buf.dtype),
            int(op),
        )
    )
    return buf


def signal_error() -> None:
    """Kill the process."""
    _check_call(_LIB.XGCommunicatorSignalError())


def _find_nccl() -> Optional[str]:
    from nvidia.nccl import lib

    # There are two versions of nvidia-nccl, one is from PyPI, another one from
    # nvidia-pyindex. We support only the first one as the second one is too old (2.9.8
    # as of writing).
    #
    # nccl 2.28 doesn't have the __file__ attribute, we use the namespace path instead.
    if lib.__file__ is not None:
        dirname: Optional[str] = os.path.dirname(lib.__file__)
    elif hasattr(lib, "__path__") and len(lib.__path__) > 0:
        dirname = lib.__path__[0]
    else:
        dirname = None
    if not dirname:
        return None

    # Find the first shared object in the lib directory.
    files = os.listdir(dirname)
    if not files:
        return None

    libname: Optional[str] = None
    for name in files:
        if name.startswith("libnccl.so"):
            libname = name
            break

    if libname is not None:
        path = os.path.join(dirname, libname)
        return path
    return None


class CommunicatorContext:
    """A context controlling collective communicator initialization and finalization."""

    def __init__(self, **args: _ArgVals) -> None:
        self.args = args
        key = "dmlc_nccl_path"
        if args.get(key, None) is not None:
            return

        binfo = build_info()
        if not binfo["USE_DLOPEN_NCCL"]:
            return

        try:
            # PyPI package of NCCL.
            path = _find_nccl()
            if path:
                self.args[key] = path
        except ImportError:
            pass

    def __enter__(self) -> _Args:
        init(**self.args)
        assert is_distributed()
        LOGGER.debug("-------------- communicator say hello ------------------")
        return self.args

    def __exit__(self, *args: Any) -> None:
        finalize()
        LOGGER.debug("--------------- communicator say bye ------------------")



"""
Getting started with learning to rank
=====================================

  .. versionadded:: 2.0.0

This is a demonstration of using XGBoost for learning to rank tasks using the
MSLR_10k_letor dataset. For more infomation about the dataset, please visit its
`description page <https://www.microsoft.com/en-us/research/project/mslr/>`_.

This is a two-part demo, the first one contains a basic example of using XGBoost to
train on relevance degree, and the second part simulates click data and enable the
position debiasing training.

For an overview of learning to rank in XGBoost, please see :doc:`Learning to Rank
</tutorials/learning_to_rank>`.
"""

from __future__ import annotations

import argparse
import json
import os
import pickle as pkl

import numpy as np
import pandas as pd
from sklearn.datasets import load_svmlight_file

import xgboost as xgb
from xgboost.testing.data import RelDataCV, simulate_clicks, sort_ltr_samples


def load_mslr_10k(data_path: str, cache_path: str) -> RelDataCV:
    """Load the MSLR10k dataset from data_path and cache a pickle object in cache_path.

    Returns
    -------

    A list of tuples [(X, y, qid), ...].

    """
    root_path = os.path.expanduser(args.data)
    cacheroot_path = os.path.expanduser(args.cache)
    cache_path = os.path.join(cacheroot_path, "MSLR_10K_LETOR.pkl")

    # Use only the Fold1 for demo:
    # Train,      Valid, Test
    # {S1,S2,S3}, S4,    S5
    fold = 1

    if not os.path.exists(cache_path):
        fold_path = os.path.join(root_path, f"Fold{fold}")
        train_path = os.path.join(fold_path, "train.txt")
        valid_path = os.path.join(fold_path, "vali.txt")
        test_path = os.path.join(fold_path, "test.txt")
        X_train, y_train, qid_train = load_svmlight_file(
            train_path, query_id=True, dtype=np.float32
        )
        y_train = y_train.astype(np.int32)
        qid_train = qid_train.astype(np.int32)

        X_valid, y_valid, qid_valid = load_svmlight_file(
            valid_path, query_id=True, dtype=np.float32
        )
        y_valid = y_valid.astype(np.int32)
        qid_valid = qid_valid.astype(np.int32)

        X_test, y_test, qid_test = load_svmlight_file(
            test_path, query_id=True, dtype=np.float32
        )
        y_test = y_test.astype(np.int32)
        qid_test = qid_test.astype(np.int32)

        data = RelDataCV(
            train=(X_train, y_train, qid_train),
            test=(X_test, y_test, qid_test),
            max_rel=4,
        )

        with open(cache_path, "wb") as fd:
            pkl.dump(data, fd)

    with open(cache_path, "rb") as fd:
        data = pkl.load(fd)

    return data


def ranking_demo(args: argparse.Namespace) -> None:
    """Demonstration for learning to rank with relevance degree."""
    data = load_mslr_10k(args.data, args.cache)

    # Sort data according to query index
    X_train, y_train, qid_train = data.train
    sorted_idx = np.argsort(qid_train)
    X_train = X_train[sorted_idx]
    y_train = y_train[sorted_idx]
    qid_train = qid_train[sorted_idx]

    X_test, y_test, qid_test = data.test
    sorted_idx = np.argsort(qid_test)
    X_test = X_test[sorted_idx]
    y_test = y_test[sorted_idx]
    qid_test = qid_test[sorted_idx]

    ranker = xgb.XGBRanker(
        tree_method="hist",
        device="cuda",
        lambdarank_pair_method="topk",
        lambdarank_num_pair_per_sample=13,
        eval_metric=["ndcg@1", "ndcg@8"],
    )
    ranker.fit(
        X_train,
        y_train,
        qid=qid_train,
        eval_set=[(X_test, y_test)],
        eval_qid=[qid_test],
        verbose=True,
    )


def click_data_demo(args: argparse.Namespace) -> None:
    """Demonstration for learning to rank with click data."""
    data = load_mslr_10k(args.data, args.cache)
    train, test = simulate_clicks(data)
    assert test is not None

    assert train.X.shape[0] == train.click.size
    assert test.X.shape[0] == test.click.size
    assert test.score.dtype == np.float32
    assert test.click.dtype == np.int32

    X_train, clicks_train, y_train, qid_train = sort_ltr_samples(
        train.X,
        train.y,
        train.qid,
        train.click,
        train.pos,
    )
    X_test, clicks_test, y_test, qid_test = sort_ltr_samples(
        test.X,
        test.y,
        test.qid,
        test.click,
        test.pos,
    )

    class ShowPosition(xgb.callback.TrainingCallback):
        def after_iteration(
            self,
            model: xgb.Booster,
            epoch: int,
            evals_log: xgb.callback.TrainingCallback.EvalsLog,
        ) -> bool:
            config = json.loads(model.save_config())
            ti_plus = np.array(config["learner"]["objective"]["ti+"])
            tj_minus = np.array(config["learner"]["objective"]["tj-"])
            df = pd.DataFrame({"ti+": ti_plus, "tj-": tj_minus})
            print(df)
            return False

    ranker = xgb.XGBRanker(
        n_estimators=512,
        tree_method="hist",
        device="cuda",
        learning_rate=0.01,
        reg_lambda=1.5,
        subsample=0.8,
        sampling_method="gradient_based",
        # LTR specific parameters
        objective="rank:ndcg",
        # - Enable bias estimation
        lambdarank_unbiased=True,
        # - normalization (1 / (norm + 1))
        lambdarank_bias_norm=1,
        # - Focus on the top 12 documents
        lambdarank_num_pair_per_sample=12,
        lambdarank_pair_method="topk",
        ndcg_exp_gain=True,
        eval_metric=["ndcg@1", "ndcg@3", "ndcg@5", "ndcg@10"],
        callbacks=[ShowPosition()],
    )
    ranker.fit(
        X_train,
        clicks_train,
        qid=qid_train,
        eval_set=[(X_test, y_test), (X_test, clicks_test)],
        eval_qid=[qid_test, qid_test],
        verbose=True,
    )
    ranker.predict(X_test)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Demonstration of learning to rank using XGBoost."
    )
    parser.add_argument(
        "--data",
        type=str,
        help="Root directory of the MSLR-WEB10K data.",
        required=True,
    )
    parser.add_argument(
        "--cache",
        type=str,
        help="Directory for caching processed data.",
        required=True,
    )
    args = parser.parse_args()

    ranking_demo(args)
    click_data_demo(args)



"""Copyright 2019-2026, XGBoost contributors"""

import asyncio
import json
from collections import OrderedDict
from inspect import signature
from pathlib import Path
from typing import Any, Dict, List, Type, TypeVar

import numpy as np
import pytest
import xgboost as xgb
from hypothesis import given, note, settings, strategies
from hypothesis._settings import duration
from packaging.version import parse as parse_version
from xgboost import testing as tm
from xgboost.collective import CommunicatorContext
from xgboost.testing.dask import get_rabit_args, make_categorical, run_recode
from xgboost.testing.params import hist_parameter_strategy

from ..test_with_dask.test_with_dask import (
    generate_array,
    run_auc,
    run_boost_from_prediction,
    run_boost_from_prediction_multi_class,
    run_categorical,
    run_dask_classifier,
    run_empty_dmatrix_auc,
    run_empty_dmatrix_cls,
    run_empty_dmatrix_reg,
    run_tree_stats,
    suppress,
)
from ..test_with_dask.test_with_dask import kCols as random_cols

pytestmark = [
    pytest.mark.skipif(**tm.no_dask()),
    pytest.mark.skipif(**tm.no_dask_cuda()),
    tm.timeout(180),
]

import cudf
import dask
import dask.dataframe as dd
from dask import __version__ as dask_version
from dask import array as da
from dask.distributed import Client
from dask_cuda import LocalCUDACluster
from xgboost import dask as dxgb
from xgboost.testing.dask import check_init_estimation, check_uneven_nan

dask_version_ge110 = dask_version and parse_version(dask_version) >= parse_version(
    "2024.11.0"
)


def run_with_dask_dataframe(DMatrixT: Type, client: Client) -> None:
    import cupy as cp

    cp.cuda.runtime.setDevice(0)
    _X, _y, _ = generate_array()

    X = dd.from_dask_array(_X).to_backend("cudf")
    y = dd.from_dask_array(_y).to_backend("cudf")

    dtrain = DMatrixT(client, X, y)
    out = dxgb.train(
        client,
        {"tree_method": "hist", "debug_synchronize": True, "device": "cuda"},
        dtrain=dtrain,
        evals=[(dtrain, "X")],
        num_boost_round=4,
    )

    assert isinstance(out["booster"], dxgb.Booster)
    assert len(out["history"]["X"]["rmse"]) == 4

    predictions = dxgb.predict(client, out, dtrain)
    assert isinstance(predictions.compute(), np.ndarray)

    series_predictions = dxgb.inplace_predict(client, out, X)
    assert isinstance(series_predictions, dd.Series)

    single_node = out["booster"].predict(xgb.DMatrix(X.compute()))

    cp.testing.assert_allclose(single_node, predictions.compute())
    np.testing.assert_allclose(single_node, series_predictions.compute().to_numpy())

    predt = dxgb.predict(client, out, X)
    assert isinstance(predt, dd.Series)

    T = TypeVar("T")

    def is_df(part: T) -> T:
        assert isinstance(part, cudf.DataFrame), part
        return part

    predt.map_partitions(is_df, meta=dd.utils.make_meta({"prediction": "f4"}))

    cp.testing.assert_allclose(predt.values.compute(), single_node)

    # Work around https://github.com/dmlc/xgboost/issues/10752
    X.columns = X.columns.astype("object")
    # Make sure the output can be integrated back to original dataframe
    X["predict"] = predictions
    X["inplace_predict"] = series_predictions

    has_null = X.isnull().values.any().compute()
    assert bool(has_null) is False


def run_with_dask_array(DMatrixT: Type, client: Client) -> None:
    import cupy as cp

    cp.cuda.runtime.setDevice(0)
    X, y, _ = generate_array()

    X = X.map_blocks(cp.asarray)  # type: ignore
    y = y.map_blocks(cp.asarray)  # type: ignore
    dtrain = DMatrixT(client, X, y)
    out = dxgb.train(
        client,
        {"tree_method": "hist", "debug_synchronize": True, "device": "cuda"},
        dtrain=dtrain,
        evals=[(dtrain, "X")],
        num_boost_round=2,
    )
    from_dmatrix = dxgb.predict(client, out, dtrain).compute()
    assert (
        json.loads(out["booster"].save_config())["learner"]["gradient_booster"][
            "updater"
        ][0]["name"]
        == "grow_gpu_hist"
    )
    inplace_predictions = dxgb.inplace_predict(client, out, X).compute()
    single_node = out["booster"].predict(xgb.DMatrix(X.compute()))
    np.testing.assert_allclose(single_node, from_dmatrix)
    device = cp.cuda.runtime.getDevice()
    assert device == inplace_predictions.device.id
    single_node = cp.array(single_node)
    assert device == single_node.device.id
    cp.testing.assert_allclose(single_node, inplace_predictions)


def to_cp(x: Any, DMatrixT: Type) -> Any:
    import cupy

    if isinstance(x, np.ndarray) and DMatrixT is dxgb.DaskQuantileDMatrix:
        X = cupy.array(x)
    else:
        X = x
    return X


def run_gpu_hist(
    params: Dict,
    num_rounds: int,
    dataset: tm.TestDataset,
    DMatrixT: Type,
    client: Client,
) -> None:
    params["device"] = "cuda"
    params = dataset.set_params(params)
    # It doesn't make sense to distribute a completely
    # empty dataset.
    if dataset.X.shape[0] == 0:
        return

    chunk = 128
    X = to_cp(dataset.X, DMatrixT)
    X = da.from_array(X, chunks=(chunk, dataset.X.shape[1]))
    y = to_cp(dataset.y, DMatrixT)
    y_chunk = chunk if len(dataset.y.shape) == 1 else (chunk, dataset.y.shape[1])
    y = da.from_array(y, chunks=y_chunk)

    if dataset.w is not None:
        w = to_cp(dataset.w, DMatrixT)
        w = da.from_array(w, chunks=(chunk,))
    else:
        w = None

    if DMatrixT is dxgb.DaskQuantileDMatrix:
        m = DMatrixT(
            client, data=X, label=y, weight=w, max_bin=params.get("max_bin", 256)
        )
    else:
        m = DMatrixT(client, data=X, label=y, weight=w)
    history = dxgb.train(
        client,
        params=params,
        dtrain=m,
        num_boost_round=num_rounds,
        evals=[(m, "train")],
    )["history"]["train"][dataset.metric]
    note(str(history))

    # See note on `ObjFunction::UpdateTreeLeaf`.
    update_leaf = dataset.name.endswith("-l1")
    if update_leaf:
        assert history[0] + 1e-2 >= history[-1]
        return
    else:
        assert tm.non_increasing(history)


def test_tree_stats() -> None:
    with LocalCUDACluster(n_workers=1) as cluster:
        with Client(cluster) as client:
            local = run_tree_stats(client, "hist", "cuda")

    with LocalCUDACluster(n_workers=2) as cluster:
        with Client(cluster) as client:
            distributed = run_tree_stats(client, "hist", "cuda")

    assert local == distributed


class TestDistributedGPU:
    @pytest.mark.skipif(**tm.no_cudf())
    def test_boost_from_prediction(self, local_cuda_client: Client) -> None:
        from sklearn.datasets import load_breast_cancer, load_iris

        X_, y_ = load_breast_cancer(return_X_y=True)
        X = dd.from_array(X_, chunksize=100).to_backend("cudf")
        y = dd.from_array(y_, chunksize=100).to_backend("cudf")
        run_boost_from_prediction(X, y, "hist", "cuda", local_cuda_client)

        X_, y_ = load_iris(return_X_y=True)
        X = dd.from_array(X_, chunksize=50).to_backend("cudf")
        y = dd.from_array(y_, chunksize=50).to_backend("cudf")
        run_boost_from_prediction_multi_class(X, y, "hist", "cuda", local_cuda_client)

    def test_init_estimation(self, local_cuda_client: Client) -> None:
        check_init_estimation("hist", "cuda", local_cuda_client)

    def test_uneven_nan(self) -> None:
        n_workers = 2
        with LocalCUDACluster(n_workers=n_workers) as cluster:
            with Client(cluster) as client:
                check_uneven_nan(client, "hist", "cuda", n_workers)

    @pytest.mark.skipif(**tm.no_dask_cudf())
    @pytest.mark.xfail(reason="Incompatible with Dask 2025.2.0+")
    def test_dask_dataframe(self, local_cuda_client: Client) -> None:
        run_with_dask_dataframe(dxgb.DaskDMatrix, local_cuda_client)
        run_with_dask_dataframe(dxgb.DaskQuantileDMatrix, local_cuda_client)

    @given(
        params=hist_parameter_strategy,
        num_rounds=strategies.integers(1, 20),
        dataset=tm.make_dataset_strategy(),
        dmatrix_type=strategies.sampled_from(
            [dxgb.DaskDMatrix, dxgb.DaskQuantileDMatrix]
        ),
    )
    @settings(
        deadline=duration(seconds=120),
        max_examples=20,
        suppress_health_check=suppress,
        print_blob=True,
    )
    @pytest.mark.skipif(**tm.no_cupy())
    def test_gpu_hist(
        self,
        params: Dict,
        num_rounds: int,
        dataset: tm.TestDataset,
        dmatrix_type: type,
        local_cuda_client: Client,
    ) -> None:
        params["tree_method"] = "hist"
        run_gpu_hist(params, num_rounds, dataset, dmatrix_type, local_cuda_client)

    @given(
        params=hist_parameter_strategy,
        num_rounds=strategies.integers(1, 20),
        dataset=tm.make_dataset_strategy(),
    )
    @settings(
        deadline=duration(seconds=120),
        max_examples=20,
        suppress_health_check=suppress,
        print_blob=True,
    )
    @pytest.mark.skipif(**tm.no_cupy())
    def test_gpu_approx(
        self,
        params: Dict,
        num_rounds: int,
        dataset: tm.TestDataset,
        local_cuda_client: Client,
    ) -> None:
        params["tree_method"] = "approx"
        run_gpu_hist(params, num_rounds, dataset, dxgb.DaskDMatrix, local_cuda_client)

    def test_empty_quantile_dmatrix(self, local_cuda_client: Client) -> None:
        client = local_cuda_client

        X, y = make_categorical(client, 1, 30, 13)
        X_valid, y_valid = make_categorical(client, 10000, 30, 13)

        Xy = dxgb.DaskQuantileDMatrix(client, X, y)
        Xy_valid = dxgb.DaskQuantileDMatrix(client, X_valid, y_valid, ref=Xy)
        # The error is from a worker. Dask cannot prioritize which worker's error to
        # propagate, it could be the emtpy DMatrix error or the collective communication
        # error. As a result, the test doesn't match the error message.
        with pytest.raises(ValueError):
            dxgb.train(
                client,
                {"tree_method": "hist", "device": "cuda", "debug_synchronize": True},
                Xy,
                num_boost_round=10,
                evals=[(Xy_valid, "Valid")],
            )

    @pytest.mark.skipif(**tm.no_cupy())
    def test_dask_array(self, local_cuda_client: Client) -> None:
        run_with_dask_array(dxgb.DaskDMatrix, local_cuda_client)
        run_with_dask_array(dxgb.DaskQuantileDMatrix, local_cuda_client)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_early_stopping(self, local_cuda_client: Client) -> None:
        from sklearn.datasets import load_breast_cancer

        X, y = load_breast_cancer(return_X_y=True)
        X, y = da.from_array(X), da.from_array(y)

        m = dxgb.DaskDMatrix(local_cuda_client, X, y)

        valid = dxgb.DaskDMatrix(local_cuda_client, X, y)
        early_stopping_rounds = 5
        booster = dxgb.train(
            local_cuda_client,
            {
                "objective": "binary:logistic",
                "eval_metric": "error",
                "tree_method": "hist",
                "device": "cuda",
            },
            m,
            evals=[(valid, "Valid")],
            num_boost_round=1000,
            early_stopping_rounds=early_stopping_rounds,
        )["booster"]
        assert hasattr(booster, "best_score")
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

        valid_X = X
        valid_y = y
        cls = dxgb.DaskXGBClassifier(
            objective="binary:logistic",
            tree_method="hist",
            device="cuda",
            eval_metric="error",
            n_estimators=100,
            early_stopping_rounds=early_stopping_rounds,
        )
        cls.client = local_cuda_client
        cls.fit(
            X,
            y,
            eval_set=[(valid_X, valid_y)],
        )
        booster = cls.get_booster()
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

    @pytest.mark.xfail(
        dask_version_ge110, reason="Test cannot pass with Dask 2024.11.0+"
    )
    @pytest.mark.skipif(**tm.no_cudf())
    @pytest.mark.parametrize("model", ["boosting"])
    def test_dask_classifier(self, model: str, local_cuda_client: Client) -> None:
        X_, y_, w_ = generate_array(with_weights=True)
        y_ = (y_ * 10).astype(np.int32)
        X = dd.from_dask_array(X_).to_backend("cudf")
        y = dd.from_dask_array(y_).to_backend("cudf")
        w = dd.from_dask_array(w_).to_backend("cudf")
        run_dask_classifier(X, y, w, model, "hist", "cuda", local_cuda_client, 10)

    def test_empty_dmatrix(self, local_cuda_client: Client) -> None:
        parameters = {
            "tree_method": "hist",
            "debug_synchronize": True,
            "device": "cuda",
        }
        run_empty_dmatrix_reg(local_cuda_client, parameters)
        run_empty_dmatrix_cls(local_cuda_client, parameters)

    @pytest.mark.skipif(**tm.no_dask_cudf())
    def test_empty_partition(self, local_cuda_client: Client) -> None:
        import cudf
        import cupy
        import dask_cudf

        mult = 100
        df = cudf.DataFrame(
            {
                "a": [1, 2, 3, 4, 5.1] * mult,
                "b": [10, 15, 29.3, 30, 31] * mult,
                "y": [10, 20, 30, 40.0, 50] * mult,
            }
        )
        parameters = {
            "tree_method": "hist",
            "debug_synchronize": True,
            "device": "cuda",
        }

        empty = df.iloc[:0]
        ddf = dask_cudf.concat(
            [dask_cudf.from_cudf(empty, npartitions=1)]
            + [dask_cudf.from_cudf(df, npartitions=3)]
            + [dask_cudf.from_cudf(df, npartitions=3)]
        )
        X = ddf[ddf.columns.difference(["y"])]
        y = ddf[["y"]]
        dtrain = dxgb.DaskQuantileDMatrix(local_cuda_client, X, y)
        bst_empty = dxgb.train(
            local_cuda_client, parameters, dtrain, evals=[(dtrain, "train")]
        )
        predt_empty = dxgb.predict(local_cuda_client, bst_empty, X).compute().values

        ddf = dask_cudf.concat(
            [dask_cudf.from_cudf(df, npartitions=3)]
            + [dask_cudf.from_cudf(df, npartitions=3)]
        )
        X = ddf[ddf.columns.difference(["y"])]
        y = ddf[["y"]]
        dtrain = dxgb.DaskQuantileDMatrix(local_cuda_client, X, y)
        bst = dxgb.train(
            local_cuda_client, parameters, dtrain, evals=[(dtrain, "train")]
        )

        predt = dxgb.predict(local_cuda_client, bst, X).compute().values
        cupy.testing.assert_allclose(predt, predt_empty)

        predt = dxgb.predict(local_cuda_client, bst, dtrain).compute()
        cupy.testing.assert_allclose(predt, predt_empty)

        predt = dxgb.inplace_predict(local_cuda_client, bst, X).compute().values
        cupy.testing.assert_allclose(predt, predt_empty)

        df = df.to_pandas()
        empty = df.iloc[:0]
        ddf = dd.concat(
            [dd.from_pandas(empty, npartitions=1)]
            + [dd.from_pandas(df, npartitions=3)]
            + [dd.from_pandas(df, npartitions=3)]
        )
        X = ddf[ddf.columns.difference(["y"])]
        y = ddf[["y"]]

        predt_empty = cupy.asnumpy(predt_empty)

        predt = dxgb.predict(local_cuda_client, bst_empty, X).compute().values
        np.testing.assert_allclose(predt, predt_empty)

        in_predt = (
            dxgb.inplace_predict(local_cuda_client, bst_empty, X).compute().values
        )
        np.testing.assert_allclose(predt, in_predt)

    def test_empty_dmatrix_auc(self, local_cuda_client: Client) -> None:
        n_workers = len(tm.dask.get_client_workers(local_cuda_client))
        run_empty_dmatrix_auc(local_cuda_client, "cuda", n_workers)

    def test_auc(self, local_cuda_client: Client) -> None:
        run_auc(local_cuda_client, "cuda")

    def test_invalid_ordinal(self, local_cuda_client: Client) -> None:
        """One should not specify the device ordinal with dask."""
        with pytest.raises(ValueError, match="device=cuda"):
            X, y, _ = generate_array()
            m = dxgb.DaskDMatrix(local_cuda_client, X, y)
            dxgb.train(local_cuda_client, {"device": "cuda:0"}, m)

        booster = dxgb.train(local_cuda_client, {"device": "cuda"}, m)["booster"]
        assert (
            json.loads(booster.save_config())["learner"]["generic_param"]["device"]
            == "cuda:0"
        )

    def test_data_initialization(self, local_cuda_client: Client) -> None:
        X, y, _ = generate_array()
        fw = da.random.random((random_cols,))
        fw = fw - fw.min()
        m = dxgb.DaskDMatrix(local_cuda_client, X, y, feature_weights=fw)

        workers = tm.dask.get_client_workers(local_cuda_client)
        rabit_args = get_rabit_args(local_cuda_client, len(workers))

        def worker_fn(worker_addr: str, data_ref: Dict) -> None:
            with dxgb.CommunicatorContext(**rabit_args):
                from xgboost.dask.data import _dmatrix_from_list_of_parts

                local_dtrain = _dmatrix_from_list_of_parts(
                    **data_ref, nthread=7, model=None, Xy_cats=None
                )
                fw_rows = local_dtrain.get_float_info("feature_weights").shape[0]
                assert fw_rows == local_dtrain.num_col()

        futures = []
        for i in range(len(workers)):
            futures.append(
                local_cuda_client.submit(
                    worker_fn,
                    workers[i],
                    m._create_fn_args(workers[i]),
                    pure=False,
                    workers=[workers[i]],
                )
            )
        local_cuda_client.gather(futures)

    def test_interface_consistency(self) -> None:
        """Check parameters are roughly the same between various DMatrices, with the
        same ordering.

        """

        def comp_dm_qdm(dm_names: List[str], qdm_names: List[str]) -> None:
            qdm_only = {"max_bin", "ref", "max_quantile_batches"}
            assert len(dm_names) == len(qdm_names) - len(qdm_only)
            i, j = 0, 0
            while i < len(dm_names) and j < len(qdm_names):
                if qdm_names[j] in qdm_only:
                    j += 1
                    continue
                assert dm_names[i] == qdm_names[j]
                i += 1
                j += 1

        # DaskDMatrix <-> DaskQuantileDMatrix
        sig = OrderedDict(signature(dxgb.DaskDMatrix).parameters)
        ddm_names = list(sig.keys())

        sig = OrderedDict(signature(dxgb.DaskQuantileDMatrix).parameters)
        dqdm_names = list(sig.keys())

        comp_dm_qdm(ddm_names, dqdm_names)

        # DMatrix <-> QuantileDMatrix
        sig = OrderedDict(signature(xgb.DMatrix).parameters)
        dm_names = list(sig.keys())
        sig = OrderedDict(signature(xgb.QuantileDMatrix).parameters)
        qdm_names = list(sig.keys())
        comp_dm_qdm(dm_names, qdm_names)

        def comp_dm_ddm(dm_names: List[str], ddm_names: List[str]) -> None:
            dm_only = {"nthread", "data_split_mode"}
            ddm_only = {"client"}
            assert len(dm_names) - len(dm_only) == len(ddm_names) - len(ddm_only)
            i, j = 0, 0
            while i < len(dm_names) and j < len(ddm_names):
                if dm_names[i] in dm_only:
                    i += 1
                    continue
                elif ddm_names[j] in ddm_only:
                    j += 1
                    continue
                assert dm_names[i] == ddm_names[j]
                i += 1
                j += 1

        # DaskDMatrix <-> DMatrix
        comp_dm_ddm(dm_names, ddm_names)

        # QuantileDMatrix <-> DaskQuantileDMatrix
        comp_dm_ddm(qdm_names, dqdm_names)

        sig = OrderedDict(signature(xgb.XGBRanker.fit).parameters)
        ranker_names = list(sig.keys())
        sig = OrderedDict(signature(dxgb.DaskXGBRanker.fit).parameters)
        dranker_names = list(sig.keys())

        for rn, drn in zip(ranker_names, dranker_names):
            assert rn == drn


@pytest.mark.skipif(**tm.no_dask_cudf())
def test_categorical(tmp_path: Path, local_cuda_client: Client) -> None:
    X, y = make_categorical(local_cuda_client, 10000, 30, 13)
    X = X.to_backend("cudf")

    X_onehot, _ = make_categorical(local_cuda_client, 10000, 30, 13, onehot=True)
    X_onehot = X_onehot.to_backend("cudf")
    run_categorical(local_cuda_client, "hist", "cuda", X, X_onehot, y, tmp_path)


@pytest.mark.skipif(**tm.no_dask_cudf())
def test_recode(local_cuda_client: Client) -> None:
    with dask.config.set(
        {
            "array.backend": "cupy",
            "dataframe.backend": "cudf",
        }
    ):
        run_recode(local_cuda_client, "cuda")


@pytest.mark.skipif(**tm.no_cupy())
def test_with_asyncio(local_cuda_client: Client) -> None:
    address = local_cuda_client.scheduler.address
    output = asyncio.run(run_from_dask_array_asyncio(address))
    assert isinstance(output["booster"], xgb.Booster)
    assert isinstance(output["history"], dict)


@pytest.mark.skipif(
    condition=not xgb.build_info()["USE_DLOPEN_NCCL"],
    reason="Not compiled with dlopen.",
)
def test_invalid_nccl(local_cuda_client: Client) -> None:
    client = local_cuda_client
    workers = tm.dask.get_client_workers(client)
    args = get_rabit_args(client, len(workers))

    def run(wid: int) -> None:
        ctx = CommunicatorContext(dmlc_nccl_path="foo", **args)
        X, y, w = tm.make_regression(n_samples=10, n_features=10, use_cupy=True)

        with ctx:
            with pytest.raises(ValueError, match=r"pip install"):
                xgb.QuantileDMatrix(X, y, weight=w)

    futures = client.map(run, range(len(workers)), workers=workers)
    client.gather(futures)


@pytest.mark.skipif(
    condition=not xgb.build_info()["USE_DLOPEN_NCCL"],
    reason="Not compiled with dlopen.",
)
@pytest.mark.parametrize("tree_method", ["hist", "approx"])
def test_nccl_load(local_cuda_client: Client, tree_method: str) -> None:
    X, y, w = tm.make_regression(128, 16, use_cupy=True)

    def make_model() -> None:
        xgb.XGBRegressor(
            device="cuda",
            tree_method=tree_method,
            objective="reg:quantileerror",
            verbosity=2,
            quantile_alpha=[0.2, 0.8],
        ).fit(X, y, sample_weight=w)

    # no nccl load when using single-node.
    with tm.captured_output() as (out, err):
        make_model()
        assert out.getvalue().find("NCCL") == -1
        assert err.getvalue().find("NCCL") == -1

    client = local_cuda_client
    workers = tm.dask.get_client_workers(client)
    args = get_rabit_args(client, len(workers))

    # nccl is loaded
    def run(wid: int) -> None:
        # FIXME(jiamingy): https://github.com/dmlc/xgboost/issues/9147
        from xgboost._c_api import _LIB, _register_log_callback

        _register_log_callback(_LIB)

        with CommunicatorContext(**args):
            with tm.captured_output() as (out, err):
                make_model()
                assert out.getvalue().find("Loaded shared NCCL") != -1, out.getvalue()

    futures = client.map(run, range(len(workers)), workers=workers)
    client.gather(futures)


async def run_from_dask_array_asyncio(scheduler_address: str) -> dxgb.TrainReturnT:
    async with Client(scheduler_address, asynchronous=True) as client:
        import cupy as cp

        X, y, _ = generate_array()
        X = X.to_backend("cupy")
        y = y.to_backend("cupy")

        m: dxgb.DaskDMatrix = await dxgb.DaskQuantileDMatrix(client, X, y)
        output = await dxgb.train(
            client, {"tree_method": "hist", "device": "cuda"}, dtrain=m
        )

        with_m = await dxgb.predict(client, output, m)
        with_X = await dxgb.predict(client, output, X)
        inplace = await dxgb.inplace_predict(client, output, X)
        assert isinstance(with_m, da.Array)
        assert isinstance(with_X, da.Array)
        assert isinstance(inplace, da.Array)

        cp.testing.assert_allclose(
            await client.compute(with_m), await client.compute(with_X)
        )
        cp.testing.assert_allclose(
            await client.compute(with_m), await client.compute(inplace)
        )

        client.shutdown()
        return output



"""Test plotting functions for XGBoost."""

import json

from graphviz import Source
from matplotlib.axes import Axes

from ..plotting import plot_tree, to_graphviz
from ..sklearn import XGBRegressor
from .data import make_categorical
from .utils import Device


def run_categorical(tree_method: str, device: Device) -> None:
    """Tests plotting functions for categorical features."""
    X, y = make_categorical(1000, 31, 19, onehot=False)
    reg = XGBRegressor(
        enable_categorical=True, n_estimators=10, tree_method=tree_method, device=device
    )
    reg.fit(X, y)
    trees = reg.get_booster().get_dump(dump_format="json")
    for tree in trees:
        j_tree = json.loads(tree)
        assert "leaf" in j_tree.keys() or isinstance(j_tree["split_condition"], list)

    graph = to_graphviz(reg, tree_idx=len(j_tree) - 1)
    assert isinstance(graph, Source)
    ax = plot_tree(reg, tree_idx=len(j_tree) - 1)
    assert isinstance(ax, Axes)



"""Copyright 2024, XGBoost contributors"""

import pytest
from distributed import Client, Scheduler, Worker
from distributed.utils_test import gen_cluster

from xgboost import testing as tm
from xgboost.testing.dask import check_external_memory, get_rabit_args


@pytest.mark.parametrize("is_qdm", [True, False])
@gen_cluster(client=True)
async def test_external_memory(
    client: Client, s: Scheduler, a: Worker, b: Worker, is_qdm: bool
) -> None:
    workers = tm.dask.get_client_workers(client)
    n_workers = len(workers)
    args = await get_rabit_args(client, n_workers)

    futs = client.map(
        check_external_memory,
        range(n_workers),
        n_workers=n_workers,
        device="cpu",
        comm_args=args,
        is_qdm=is_qdm,
    )
    await client.gather(futs)



"""Tests for parsing trees."""

import pytest

from ..core import DMatrix
from ..sklearn import XGBRegressor
from ..training import train
from .data import make_categorical
from .utils import Device


def run_tree_to_df_categorical(tree_method: str, device: Device) -> None:
    """Tests tree_to_df with categorical features."""
    X, y = make_categorical(100, 10, 31, onehot=False)
    Xy = DMatrix(X, y, enable_categorical=True)
    booster = train(
        {"tree_method": tree_method, "device": device}, Xy, num_boost_round=10
    )
    df = booster.trees_to_dataframe()
    for _, x in df.iterrows():
        if x["Feature"] != "Leaf":
            assert len(x["Category"]) >= 1


def run_split_value_histograms(tree_method: str, device: Device) -> None:
    """Tests split_value_histograms with categorical features."""
    X, y = make_categorical(1000, 10, 13, onehot=False)
    reg = XGBRegressor(tree_method=tree_method, enable_categorical=True, device=device)
    reg.fit(X, y)

    with pytest.raises(ValueError, match="doesn't"):
        reg.get_booster().get_split_value_histogram("3", bins=5)



"""
Demo for using cross validation
===============================
"""

import os
from typing import Any, Dict, Tuple

import numpy as np

import xgboost as xgb

# load data in do training
CURRENT_DIR = os.path.dirname(__file__)
dtrain = xgb.DMatrix(
    os.path.join(CURRENT_DIR, "../data/agaricus.txt.train?format=libsvm")
)
param = {"max_depth": 2, "eta": 1, "objective": "binary:logistic"}
num_round = 2

print("running cross validation")
# do cross validation, this will print result out as
# [iteration]  metric_name:mean_value+std_value
# std_value is standard deviation of the metric
xgb.cv(
    param,
    dtrain,
    num_round,
    nfold=5,
    metrics={"error"},
    seed=0,
    callbacks=[xgb.callback.EvaluationMonitor(show_stdv=True)],
)

print("running cross validation, disable standard deviation display")
# do cross validation, this will print result out as
# [iteration]  metric_name:mean_value
res = xgb.cv(
    param,
    dtrain,
    num_boost_round=10,
    nfold=5,
    metrics={"error"},
    seed=0,
    callbacks=[
        xgb.callback.EvaluationMonitor(show_stdv=False),
        xgb.callback.EarlyStopping(3),
    ],
)
print(res)
print("running cross validation, with preprocessing function")


# define the preprocessing function
# used to return the preprocessed training, test data, and parameter
# we can use this to do weight rescale, etc.
# as a example, we try to set scale_pos_weight
def fpreproc(
    dtrain: xgb.DMatrix, dtest: xgb.DMatrix, param: Any
) -> Tuple[xgb.DMatrix, xgb.DMatrix, Dict[str, Any]]:
    label = dtrain.get_label()
    ratio = float(np.sum(label == 0)) / np.sum(label == 1)
    param["scale_pos_weight"] = ratio
    return (dtrain, dtest, param)


# do cross validation, for each fold
# the dtrain, dtest, param will be passed into fpreproc
# then the return value of fpreproc will be used to generate
# results of that fold
xgb.cv(param, dtrain, num_round, nfold=5, metrics={"auc"}, seed=0, fpreproc=fpreproc)

###
# you can also do cross validation with customized loss function
# See custom_objective.py
##
print("running cross validation, with customized loss function")


def logregobj(preds: np.ndarray, dtrain: xgb.DMatrix) -> Tuple[np.ndarray, np.ndarray]:
    labels = dtrain.get_label()
    preds = 1.0 / (1.0 + np.exp(-preds))
    grad = preds - labels
    hess = preds * (1.0 - preds)
    return grad, hess


def evalerror(preds: np.ndarray, dtrain: xgb.DMatrix) -> Tuple[str, float]:
    labels = dtrain.get_label()
    preds = 1.0 / (1.0 + np.exp(-preds))
    return "error", float(sum(labels != (preds > 0.0))) / len(labels)


param = {"max_depth": 2, "eta": 1}
# train with customized objective
xgb.cv(
    param, dtrain, num_round, nfold=5, seed=0, obj=logregobj, custom_metric=evalerror
)



import os
import subprocess
import sys
from pathlib import Path

import pytest
import xgboost
from xgboost import testing as tm

pytestmark = tm.timeout(30)

DEMO_DIR = tm.demo_dir(__file__)
PYTHON_DEMO_DIR = os.path.join(DEMO_DIR, "guide-python")
CLI_DEMO_DIR = os.path.join(DEMO_DIR, "CLI")


PYTHON = sys.executable


def test_basic_walkthrough(tmp_path: Path) -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "basic_walkthrough.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd, cwd=tmp_path)


@pytest.mark.skipif(**tm.no_pandas())
def test_categorical() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "categorical.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_pandas())
def test_cat_pipeline() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "cat_pipeline.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_matplotlib())
def test_custom_multiclass_objective() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "custom_softmax.py")
    cmd = [PYTHON, script, "--plot=0"]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_matplotlib())
def test_custom_rmsle_objective() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "custom_rmsle.py")
    cmd = [PYTHON, script, "--plot=0"]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_matplotlib())
def test_feature_weights_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "feature_weights.py")
    cmd = [PYTHON, script, "--plot=0"]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_sklearn())
def test_sklearn_demo(tmp_path: Path) -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "sklearn_examples.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd, cwd=tmp_path)
    assert (tmp_path / "best_calif.pkl").exists()


@pytest.mark.skipif(**tm.no_sklearn())
@pytest.mark.timeout(60)
def test_sklearn_parallel_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "sklearn_parallel.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_sklearn())
def test_sklearn_evals_result_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "sklearn_evals_result.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


def test_boost_from_prediction_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "boost_from_prediction.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


def test_predict_first_ntree_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "predict_first_ntree.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


def test_individual_trees() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "individual_trees.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


def test_predict_leaf_indices_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "predict_leaf_indices.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


def test_generalized_linear_model_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "generalized_linear_model.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


def test_cross_validation_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "cross_validation.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


def test_external_memory_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "external_memory.py")
    cmd = [PYTHON, script, "--device=cpu"]
    subprocess.check_call(cmd)


def test_distributed_extmem_basic_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "distributed_extmem_basic.py")
    cmd = [PYTHON, script, "--device=cpu"]
    subprocess.check_call(cmd)


def test_evals_result_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "evals_result.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_sklearn())
@pytest.mark.skipif(**tm.no_pandas())
def test_aft_demo(tmp_path: Path) -> None:
    script = os.path.join(DEMO_DIR, "aft_survival", "aft_survival_demo.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd, cwd=tmp_path)
    assert (tmp_path / "aft_model.json").exists()


@pytest.mark.skipif(**tm.no_matplotlib())
def test_callbacks_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "callbacks.py")
    cmd = [PYTHON, script, "--plot=0"]
    subprocess.check_call(cmd)


def test_continuation_demo() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "continuation.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_sklearn())
@pytest.mark.skipif(**tm.no_matplotlib())
def test_multioutput_reg() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "multioutput_regression.py")
    cmd = [PYTHON, script, "--plot=0"]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_sklearn())
def test_prediction_intervals() -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "prediction_intervals.py")
    cmd = [PYTHON, script]
    subprocess.check_call(cmd)


@pytest.mark.skipif(**tm.no_ubjson())
def test_json_model(tmp_path: Path) -> None:
    script = os.path.join(PYTHON_DEMO_DIR, "model_parser.py")

    def run_test(reg: xgboost.XGBRegressor, suffix: str) -> None:
        path = tmp_path / f"reg_{suffix}.json"
        reg.save_model(path)
        cmd = [PYTHON, script, f"--model={path}"]
        subprocess.check_call(cmd)

        path = tmp_path / f"reg_{suffix}.ubj"
        reg.save_model(path)
        cmd = [PYTHON, script, f"--model={path}"]
        subprocess.check_call(cmd)

    # numerical
    X, y = tm.make_sparse_regression(100, 10, 0.5, False)
    reg = xgboost.XGBRegressor(n_estimators=2, tree_method="hist")
    reg.fit(X, y)
    run_test(reg, "numerical")

    # categorical
    X, y = tm.make_categorical(
        n_samples=1000,
        n_features=10,
        n_categories=6,
        onehot=False,
        sparsity=0.5,
        cat_ratio=0.5,
        shuffle=True,
    )
    reg = xgboost.XGBRegressor(n_estimators=2, tree_method="hist")
    reg.fit(X, y)
    run_test(reg, "categorical")


# - gpu_acceleration is not tested due to covertype dataset is being too huge.
# - gamma regression is not tested as it requires running a R script first.
# - aft viz is not tested due to ploting is not controlled
# - aft tunning is not tested due to extra dependency.



import numpy as np
import gc
import pytest
import xgboost as xgb
from hypothesis import given, strategies, assume, settings, note

import sys
import os

# sys.path.append("tests/python")
# import testing as tm
from xgboost import testing as tm

parameter_strategy = strategies.fixed_dictionaries(
    {
        "max_depth": strategies.integers(0, 11),
        "max_leaves": strategies.integers(0, 256),
        "max_bin": strategies.integers(2, 1024),
        "grow_policy": strategies.sampled_from(["lossguide", "depthwise"]),
        "single_precision_histogram": strategies.booleans(),
        "min_child_weight": strategies.floats(0.5, 2.0),
        "seed": strategies.integers(0, 10),
        # We cannot enable subsampling as the training loss can increase
        # 'subsample': strategies.floats(0.5, 1.0),
        "colsample_bytree": strategies.floats(0.5, 1.0),
        "colsample_bylevel": strategies.floats(0.5, 1.0),
    }
).filter(
    lambda x: (x["max_depth"] > 0 or x["max_leaves"] > 0)
    and (x["max_depth"] > 0 or x["grow_policy"] == "lossguide")
)


def train_result(param, dmat, num_rounds):
    result = {}
    xgb.train(
        param,
        dmat,
        num_rounds,
        [(dmat, "train")],
        verbose_eval=False,
        evals_result=result,
    )
    return result


class TestSYCLUpdaters:
    @given(parameter_strategy, strategies.integers(1, 5), tm.make_dataset_strategy())
    @settings(deadline=None)
    def test_sycl_hist(self, param, num_rounds, dataset):
        param["tree_method"] = "hist"
        param["device"] = "sycl"
        param["verbosity"] = 0
        param = dataset.set_params(param)
        result = train_result(param, dataset.get_dmat(), num_rounds)
        note(result)
        assert tm.non_increasing(result["train"][dataset.metric])

    @given(tm.make_dataset_strategy(), strategies.integers(0, 1))
    @settings(deadline=None)
    def test_specified_device_id_sycl_update(self, dataset, device_id):
        # Read the list of sycl-devicese
        sycl_ls = os.popen("sycl-ls").read()
        devices = sycl_ls.split("\n")

        # Test should launch only on gpu
        # Find gpus in the list of devices
        # and use the id in the list insteard of device_id
        target_device_type = "opencl:gpu"
        found_devices = 0
        for idx in range(len(devices)):
            if len(devices[idx]) >= len(target_device_type):
                if devices[idx][1 : 1 + len(target_device_type)] == target_device_type:
                    if found_devices == device_id:
                        param = {"device": f"sycl:gpu:{idx}"}
                        param = dataset.set_params(param)
                        result = train_result(param, dataset.get_dmat(), 10)
                        assert tm.non_increasing(result["train"][dataset.metric])
                    else:
                        found_devices += 1



import cupy as cp
import numpy as np
import pytest

import xgboost as xgb


# Test for integer overflow or out of memory exceptions
def test_large_input() -> None:
    available_bytes, _ = cp.cuda.runtime.memGetInfo()
    # 15 GB
    required_bytes = 1.5e10
    if available_bytes < required_bytes:
        pytest.skip("Not enough memory on this device")
    n = 1000
    m = ((1 << 31) + n - 1) // n
    assert np.log2(m * n) > 31
    X = cp.ones((m, n), dtype=np.float32)
    y = cp.ones(m)
    w = cp.ones(m)
    dmat = xgb.QuantileDMatrix(X, y, weight=w)
    booster = xgb.train(
        {"tree_method": "hist", "max_depth": 1, "device": "cuda"}, dmat, 1
    )
    del y
    booster.inplace_predict(X)



"""
Example of training with Dask on GPU
====================================
"""

import dask
import dask_cudf
from dask import array as da
from dask import dataframe as dd
from dask.distributed import Client
from dask_cuda import LocalCUDACluster

from xgboost import dask as dxgb
from xgboost.dask import DaskDMatrix


def using_dask_matrix(client: Client, X: da.Array, y: da.Array) -> da.Array:
    # DaskDMatrix acts like normal DMatrix, works as a proxy for local DMatrix scatter
    # around workers.
    dtrain = DaskDMatrix(client, X, y)

    # Use train method from xgboost.dask instead of xgboost.  This distributed version
    # of train returns a dictionary containing the resulting booster and evaluation
    # history obtained from evaluation metrics.
    output = dxgb.train(
        client,
        # Make sure the device is set to CUDA.
        {"tree_method": "hist", "device": "cuda"},
        dtrain,
        num_boost_round=4,
        evals=[(dtrain, "train")],
    )
    bst = output["booster"]
    history = output["history"]

    # you can pass output directly into `predict` too.
    prediction = dxgb.predict(client, bst, dtrain)
    print("Evaluation history:", history)
    return prediction


def using_quantile_device_dmatrix(client: Client, X: da.Array, y: da.Array) -> da.Array:
    """`DaskQuantileDMatrix` is a data type specialized for `hist` tree methods for
     reducing memory usage.

    .. versionadded:: 1.2.0

    """
    # `DaskQuantileDMatrix` is used instead of `DaskDMatrix`, be careful that it can not
    # be used for anything else other than training unless a reference is specified. See
    # the `ref` argument of `DaskQuantileDMatrix`.
    dtrain = dxgb.DaskQuantileDMatrix(client, X, y)
    output = dxgb.train(
        client,
        # Make sure the device is set to CUDA.
        {"tree_method": "hist", "device": "cuda"},
        dtrain,
        num_boost_round=4,
        evals=[(dtrain, "train")],
    )

    prediction = dxgb.predict(client, output, X)
    return prediction


if __name__ == "__main__":
    # `LocalCUDACluster` is used for assigning GPU to XGBoost processes.  Here
    # `n_workers` represents the number of GPUs since we use one GPU per worker process.
    with LocalCUDACluster(n_workers=2, threads_per_worker=4) as cluster:
        # Create client from cluster, set the backend to GPU array (cupy).
        with Client(cluster) as client, dask.config.set({"array.backend": "cupy"}):
            # Generate some random data for demonstration
            rng = da.random.default_rng(1)

            m = 2**18
            n = 100
            X = rng.uniform(size=(m, n), chunks=(128**2, -1))
            y = X.sum(axis=1)

            X = dd.from_dask_array(X)
            y = dd.from_dask_array(y)
            # XGBoost can take arrays. This is to show that DataFrame uses the GPU
            # backend as well.
            assert isinstance(X, dask_cudf.DataFrame)
            assert isinstance(y, dask_cudf.Series)

            print("Using DaskQuantileDMatrix")
            from_ddqdm = using_quantile_device_dmatrix(client, X, y).compute()
            print("Using DMatrix")
            from_dmatrix = using_dask_matrix(client, X, y).compute()



import numpy as np
import pytest

import xgboost as xgb
from xgboost.testing.data import get_california_housing

try:
    import shap
except Exception:
    shap = None
    pass


pytestmark = pytest.mark.skipif(shap is None, reason="Requires shap package")


# xgboost removed ntree_limit in 2.0, which breaks the SHAP package.
@pytest.mark.xfail
def test_with_shap() -> None:
    X, y = get_california_housing()
    dtrain = xgb.DMatrix(X, label=y)
    model = xgb.train({"learning_rate": 0.01}, dtrain, 10)
    explainer = shap.TreeExplainer(model)
    shap_values = explainer.shap_values(X)
    margin = model.predict(dtrain, output_margin=True)
    assert np.allclose(
        np.sum(shap_values, axis=len(shap_values.shape) - 1),
        margin - explainer.expected_value,
        1e-3,
        1e-3,
    )



import unittest

import pytest
from xgboost import testing as tm

pytestmark = [pytest.mark.skipif(**tm.no_spark())]

from xgboost.spark.utils import _get_default_params_from_func


class UtilsTest(unittest.TestCase):
    def test_get_default_params(self):
        class Foo:
            def func1(self, x, y, key1=None, key2="val2", key3=0, key4=None):
                pass

        unsupported_params = {"key2", "key4"}
        expected_default_params = {
            "key1": None,
            "key3": 0,
        }
        actual_default_params = _get_default_params_from_func(
            Foo.func1, unsupported_params
        )
        self.assertEqual(
            len(expected_default_params.keys()), len(actual_default_params.keys())
        )
        for k, v in actual_default_params.items():
            self.assertEqual(expected_default_params[k], v)



# pylint: disable=too-many-arguments, too-many-branches, too-many-lines
# pylint: disable=too-many-return-statements
"""Data dispatching for DMatrix."""

import ctypes
import functools
import json
import os
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeAlias,
    TypeGuard,
    Union,
)

import numpy as np

from ._c_api import _LIB, _check_call, c_str, make_jcargs
from ._data_utils import (
    AifType,
    Categories,
    DfCatAccessor,
    TransformedDf,
    _arrow_array_inf,
    _ensure_np_dtype,
    _is_df_cat,
    array_hasobject,
    array_interface,
    array_interface_dict,
    arrow_cat_inf,
    check_cudf_meta,
    cuda_array_interface,
    cuda_array_interface_dict,
    cudf_cat_inf,
    get_ref_categories,
    is_arrow_dict,
    pd_cat_inf,
)
from ._typing import (
    CupyT,
    DataSplitMode,
    DataType,
    FeatureNames,
    FeatureTypes,
    FloatCompatible,
    NumpyDType,
    PandasDType,
    PathLike,
    TransformedData,
    c_bst_ulong,
)
from .compat import (
    _is_arrow,
    _is_cudf_df,
    _is_cudf_pandas,
    _is_cudf_ser,
    _is_cupy_alike,
    _is_modin_df,
    _is_modin_series,
    _is_pandas_df,
    _is_pandas_series,
    _is_polars,
    _is_polars_lazyframe,
    _is_polars_series,
    import_pandas,
    import_polars,
    import_pyarrow,
    is_pyarrow_available,
    lazy_isinstance,
)

if TYPE_CHECKING:
    import pyarrow as pa
    from pandas import DataFrame as PdDataFrame
    from pandas import Series as PdSeries

    from .core import DMatrix, _ProxyDMatrix


DispatchedDataBackendReturnType: TypeAlias = Tuple[
    ctypes.c_void_p, Optional[FeatureNames], Optional[FeatureTypes]
]

CAT_T = "c"

# meta info that can be a matrix instead of vector.
_matrix_meta = {"base_margin", "label"}


def _warn_unused_missing(data: DataType, missing: Optional[FloatCompatible]) -> None:
    if (missing is not None) and (not np.isnan(missing)):
        warnings.warn(
            "`missing` is not used for current input data type:" + str(type(data)),
            UserWarning,
        )


def _check_data_shape(data: DataType) -> None:
    if hasattr(data, "shape") and len(data.shape) != 2:
        raise ValueError("Please reshape the input data into 2-dimensional matrix.")


def is_scipy_csr(data: DataType) -> bool:
    """Predicate for scipy CSR input."""
    is_array = False
    is_matrix = False
    try:
        from scipy.sparse import csr_array

        is_array = isinstance(data, csr_array)
    except ImportError:
        pass
    try:
        from scipy.sparse import csr_matrix

        is_matrix = isinstance(data, csr_matrix)
    except ImportError:
        pass
    return is_array or is_matrix


def transform_scipy_sparse(data: DataType, is_csr: bool) -> DataType:
    """Ensure correct data alignment and data type for scipy sparse inputs. Input should
    be either csr or csc matrix.

    """
    from scipy.sparse import csc_matrix, csr_matrix

    if len(data.indices) != len(data.data):
        raise ValueError(f"length mismatch: {len(data.indices)} vs {len(data.data)}")

    indptr, _ = _ensure_np_dtype(data.indptr, data.indptr.dtype)
    indices, _ = _ensure_np_dtype(data.indices, data.indices.dtype)
    values, _ = _ensure_np_dtype(data.data, data.data.dtype)
    if (
        indptr is not data.indptr
        or indices is not data.indices
        or values is not data.data
    ):
        if is_csr:
            data = csr_matrix((values, indices, indptr), shape=data.shape)
        else:
            data = csc_matrix((values, indices, indptr), shape=data.shape)
    return data


def _from_scipy_csr(
    *,
    data: DataType,
    missing: FloatCompatible,
    nthread: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    """Initialize data from a CSR matrix."""

    handle = ctypes.c_void_p()
    data = transform_scipy_sparse(data, True)
    _check_call(
        _LIB.XGDMatrixCreateFromCSR(
            array_interface(data.indptr),
            array_interface(data.indices),
            array_interface(data.data),
            c_bst_ulong(data.shape[1]),
            make_jcargs(
                missing=float(missing),
                nthread=int(nthread),
                data_split_mode=int(data_split_mode),
            ),
            ctypes.byref(handle),
        )
    )
    return handle, feature_names, feature_types


def is_scipy_csc(data: DataType) -> bool:
    """Predicate for scipy CSC input."""
    is_array = False
    is_matrix = False
    try:
        from scipy.sparse import csc_array

        is_array = isinstance(data, csc_array)
    except ImportError:
        pass
    try:
        from scipy.sparse import csc_matrix

        is_matrix = isinstance(data, csc_matrix)
    except ImportError:
        pass
    return is_array or is_matrix


def _from_scipy_csc(
    *,
    data: DataType,
    missing: FloatCompatible,
    nthread: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    """Initialize data from a CSC matrix."""
    handle = ctypes.c_void_p()
    transform_scipy_sparse(data, False)
    _check_call(
        _LIB.XGDMatrixCreateFromCSC(
            array_interface(data.indptr),
            array_interface(data.indices),
            array_interface(data.data),
            c_bst_ulong(data.shape[0]),
            make_jcargs(
                missing=float(missing),
                nthread=int(nthread),
                data_split_mode=int(data_split_mode),
            ),
            ctypes.byref(handle),
        )
    )
    return handle, feature_names, feature_types


def is_scipy_coo(data: DataType) -> bool:
    """Predicate for scipy COO input."""
    is_array = False
    is_matrix = False
    try:
        from scipy.sparse import coo_array

        is_array = isinstance(data, coo_array)
    except ImportError:
        pass
    try:
        from scipy.sparse import coo_matrix

        is_matrix = isinstance(data, coo_matrix)
    except ImportError:
        pass
    return is_array or is_matrix


def _is_np_array_like(data: DataType) -> TypeGuard[np.ndarray]:
    return hasattr(data, "__array_interface__")


def _maybe_np_slice(data: DataType, dtype: Optional[NumpyDType]) -> np.ndarray:
    """Handle numpy slice.  This can be removed if we use __array_interface__."""
    try:
        if not data.flags.c_contiguous:
            data = np.array(data, copy=True, dtype=dtype)
        else:
            data = np.asarray(data, dtype=dtype)
    except AttributeError:
        data = np.asarray(data, dtype=dtype)
    data, dtype = _ensure_np_dtype(data, dtype)
    return data


def _from_numpy_array(
    *,
    data: np.ndarray,
    missing: FloatCompatible,
    nthread: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    """Initialize data from a 2-D numpy matrix."""
    _check_data_shape(data)
    data, _ = _ensure_np_dtype(data, data.dtype)
    handle = ctypes.c_void_p()
    _check_call(
        _LIB.XGDMatrixCreateFromDense(
            array_interface(data),
            make_jcargs(
                missing=float(missing),
                nthread=int(nthread),
                data_split_mode=int(data_split_mode),
            ),
            ctypes.byref(handle),
        )
    )
    return handle, feature_names, feature_types


_pandas_dtype_mapper = {
    "int8": "int",
    "int16": "int",
    "int32": "int",
    "int64": "int",
    "uint8": "int",
    "uint16": "int",
    "uint32": "int",
    "uint64": "int",
    "float16": "float",
    "float32": "float",
    "float64": "float",
    "bool": "i",
}

# nullable types
pandas_nullable_mapper = {
    "Int8": "int",
    "Int16": "int",
    "Int32": "int",
    "Int64": "int",
    "UInt8": "int",
    "UInt16": "int",
    "UInt32": "int",
    "UInt64": "int",
    "Float32": "float",
    "Float64": "float",
    "boolean": "i",
}

pandas_pyarrow_mapper = {
    "int8[pyarrow]": "int",
    "int16[pyarrow]": "int",
    "int32[pyarrow]": "int",
    "int64[pyarrow]": "int",
    "uint8[pyarrow]": "int",
    "uint16[pyarrow]": "int",
    "uint32[pyarrow]": "int",
    "uint64[pyarrow]": "int",
    "float[pyarrow]": "float",
    "float32[pyarrow]": "float",
    "double[pyarrow]": "float",
    "float64[pyarrow]": "float",
    "bool[pyarrow]": "i",
}

_pandas_dtype_mapper.update(pandas_nullable_mapper)
_pandas_dtype_mapper.update(pandas_pyarrow_mapper)


_ENABLE_CAT_ERR = (
    "When categorical type is supplied, the experimental DMatrix parameter"
    "`enable_categorical` must be set to `True`."
)


def _invalid_dataframe_dtype(data: DataType) -> None:
    # pandas series has `dtypes` but it's just a single object
    # cudf series doesn't have `dtypes`.
    if hasattr(data, "dtypes") and hasattr(data.dtypes, "__iter__"):
        bad_fields = [
            f"{data.columns[i]}: {dtype}"
            for i, dtype in enumerate(data.dtypes)
            if dtype.name not in _pandas_dtype_mapper
        ]
        err = " Invalid columns:" + ", ".join(bad_fields)
    else:
        err = ""

    type_err = "DataFrame.dtypes for data must be int, float, bool or category."
    msg = f"""{type_err} {_ENABLE_CAT_ERR} {err}"""
    raise ValueError(msg)


def pandas_feature_info(
    data: "PdDataFrame",
    meta: Optional[str],
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    enable_categorical: bool,
) -> Tuple[Optional[FeatureNames], Optional[FeatureTypes]]:
    """Handle feature info for pandas dataframe."""
    pd = import_pandas()

    # handle feature names
    if feature_names is None and meta is None:
        if isinstance(data.columns, pd.MultiIndex):
            feature_names = [" ".join([str(x) for x in i]) for i in data.columns]
        else:
            feature_names = list(data.columns.map(str))

    # handle feature types and dtype validation
    new_feature_types = []
    need_sparse_extension_warn = True
    for dtype in data.dtypes:
        if is_pd_sparse_dtype(dtype):
            new_feature_types.append(_pandas_dtype_mapper[dtype.subtype.name])
            if need_sparse_extension_warn:
                warnings.warn("Sparse arrays from pandas are converted into dense.")
                need_sparse_extension_warn = False
        elif (
            is_pd_cat_dtype(dtype) or is_pa_ext_categorical_dtype(dtype)
        ) and enable_categorical:
            new_feature_types.append(CAT_T)
        else:
            try:
                new_feature_types.append(_pandas_dtype_mapper[dtype.name])
            except KeyError:
                _invalid_dataframe_dtype(data)

    if feature_types is None and meta is None:
        feature_types = new_feature_types

    return feature_names, feature_types


def is_nullable_dtype(dtype: PandasDType) -> bool:
    """Whether dtype is a pandas nullable type."""

    from pandas.api.extensions import ExtensionDtype

    if not isinstance(dtype, ExtensionDtype):
        return False

    from pandas.api.types import is_bool_dtype, is_float_dtype, is_integer_dtype

    is_int = is_integer_dtype(dtype) and dtype.name in pandas_nullable_mapper
    # np.bool has alias `bool`, while pd.BooleanDtype has `boolean`.
    is_bool = is_bool_dtype(dtype) and dtype.name == "boolean"
    is_float = is_float_dtype(dtype) and dtype.name in pandas_nullable_mapper
    return is_int or is_bool or is_float or is_pd_cat_dtype(dtype)


def is_pa_ext_dtype(dtype: Any) -> bool:
    """Return whether dtype is a pyarrow extension type for pandas"""
    return hasattr(dtype, "pyarrow_dtype")


def is_pa_ext_categorical_dtype(dtype: Any) -> bool:
    """Check whether dtype is a dictionary type."""
    return lazy_isinstance(
        getattr(dtype, "pyarrow_dtype", None), "pyarrow.lib", "DictionaryType"
    )


@functools.cache
def _lazy_load_pd_is_cat() -> Callable[[PandasDType], bool]:
    pd = import_pandas()

    if hasattr(pd.util, "version") and hasattr(pd.util.version, "Version"):
        Version = pd.util.version.Version
        if Version(pd.__version__) >= Version("2.1.0"):
            from pandas import CategoricalDtype

            def pd_is_cat_210(dtype: PandasDType) -> bool:
                return isinstance(dtype, CategoricalDtype)

            return pd_is_cat_210
    from pandas.api.types import is_categorical_dtype  # type: ignore[attr-defined]

    return is_categorical_dtype


def is_pd_cat_dtype(dtype: PandasDType) -> bool:
    """Wrapper for testing pandas category type."""
    is_cat = _lazy_load_pd_is_cat()
    return is_cat(dtype)


@functools.cache
def _lazy_load_pd_is_sparse() -> Callable[[PandasDType], bool]:
    pd = import_pandas()

    if hasattr(pd.util, "version") and hasattr(pd.util.version, "Version"):
        Version = pd.util.version.Version
        if Version(pd.__version__) >= Version("2.1.0"):
            from pandas import SparseDtype

            def pd_is_sparse_210(dtype: PandasDType) -> bool:
                return isinstance(dtype, SparseDtype)

            return pd_is_sparse_210

    from pandas.api.types import is_sparse  # type: ignore[attr-defined]

    return is_sparse


def is_pd_sparse_dtype(dtype: PandasDType) -> bool:
    """Wrapper for testing pandas sparse type."""
    is_sparse = _lazy_load_pd_is_sparse()

    return is_sparse(dtype)


def pandas_pa_type(ser: Any) -> np.ndarray:
    """Handle pandas pyarrow extention."""
    pd = import_pandas()

    if TYPE_CHECKING:
        import pyarrow as pa
    else:
        pa = import_pyarrow()

    # No copy, callstack:
    # pandas.core.internals.managers.SingleBlockManager.array_values()
    # pandas.core.internals.blocks.EABackedBlock.values
    d_array: pd.arrays.ArrowExtensionArray = ser.array  # type: ignore[name-defined]
    # no copy in __arrow_array__
    # ArrowExtensionArray._data is a chunked array
    aa: "pa.ChunkedArray" = d_array.__arrow_array__()
    # combine_chunks takes the most significant amount of time
    chunk: "pa.Array" = aa.combine_chunks()
    # When there's null value, we have to use copy
    zero_copy = chunk.null_count == 0 and not pa.types.is_boolean(chunk.type)
    # Alternately, we can use chunk.buffers(), which returns a list of buffers and
    # we need to concatenate them ourselves.
    # FIXME(jiamingy): Is there a better way to access the arrow buffer along with
    # its mask?
    # Buffers from chunk.buffers() have the address attribute, but don't expose the
    # mask.
    arr: np.ndarray = chunk.to_numpy(zero_copy_only=zero_copy, writable=False)
    arr, _ = _ensure_np_dtype(arr, arr.dtype)
    return arr


@functools.cache
def _lazy_has_npdtypes() -> bool:
    return np.lib.NumpyVersion(np.__version__) > np.lib.NumpyVersion("1.25.0")


@functools.cache
def _lazy_load_pd_floats() -> tuple:
    from pandas import Float32Dtype, Float64Dtype

    return Float32Dtype, Float64Dtype


def pandas_transform_data(
    data: "PdDataFrame",
) -> List[Union[np.ndarray, DfCatAccessor]]:
    """Handle categorical dtype and extension types from pandas."""
    Float32Dtype, Float64Dtype = _lazy_load_pd_floats()

    result: List[Union[np.ndarray, DfCatAccessor]] = []
    np_dtypes = _lazy_has_npdtypes()

    def cat_codes(ser: "PdSeries") -> DfCatAccessor:
        return ser.cat

    def nu_type(ser: "PdSeries") -> np.ndarray:
        # Avoid conversion when possible
        if isinstance(dtype, Float32Dtype):
            res_dtype: NumpyDType = np.float32
        elif isinstance(dtype, Float64Dtype):
            res_dtype = np.float64
        else:
            res_dtype = np.float32
        return _ensure_np_dtype(
            ser.to_numpy(dtype=res_dtype, na_value=np.nan), res_dtype
        )[0]

    def oth_type(ser: "PdSeries") -> np.ndarray:
        # The dtypes module is added in 1.25.
        npdtypes = np_dtypes and isinstance(
            ser.dtype,
            (
                # pylint: disable=no-member
                np.dtypes.Float32DType,  # type: ignore[attr-defined]
                # pylint: disable=no-member
                np.dtypes.Float64DType,  # type: ignore[attr-defined]
            ),
        )

        if npdtypes or dtype in {np.float32, np.float64}:
            array = ser.to_numpy()
        else:
            # Specifying the dtype can significantly slow down the conversion (about
            # 15% slow down for dense inplace-predict)
            array = ser.to_numpy(dtype=np.float32, na_value=np.nan)
        return _ensure_np_dtype(array, array.dtype)[0]

    for col, dtype in zip(data.columns, data.dtypes):
        if is_pa_ext_categorical_dtype(dtype):
            raise ValueError(
                "pyarrow dictionary type is not supported. Use pandas category instead."
            )
        if is_pd_cat_dtype(dtype):
            result.append(cat_codes(data[col]))
        elif is_pa_ext_dtype(dtype):
            result.append(pandas_pa_type(data[col]))
        elif is_nullable_dtype(dtype):
            result.append(nu_type(data[col]))
        elif is_pd_sparse_dtype(dtype):
            arr = data[col].values
            arr = arr.to_dense()
            if _is_np_array_like(arr):
                arr, _ = _ensure_np_dtype(arr, arr.dtype)
            result.append(arr)
        else:
            result.append(oth_type(data[col]))

    # FIXME(jiamingy): Investigate the possibility of using dataframe protocol or arrow
    # IPC format for pandas so that we can apply the data transformation inside XGBoost
    # for better memory efficiency.
    return result


class PandasTransformed(TransformedDf):
    """A storage class for transformed pandas DataFrame."""

    def __init__(
        self,
        columns: List[Union[np.ndarray, DfCatAccessor]],
        ref_categories: Optional[Categories],
    ) -> None:
        self.columns = columns

        aitfs: AifType = []
        temporary_buffers = []

        # Get the array interface representation for each column.
        for col in self.columns:
            if _is_df_cat(col):
                # Categorical column
                jnames, jcodes, buf = pd_cat_inf(col.categories, col.codes)
                temporary_buffers.append(buf)
                aitfs.append((jnames, jcodes))
            else:
                assert isinstance(col, np.ndarray)
                inf = array_interface_dict(col)
                # Numeric column
                aitfs.append(inf)

        super().__init__(
            ref_categories=ref_categories,
            aitfs=aitfs,
            temporary_buffers=temporary_buffers,
        )

    @property
    def shape(self) -> Tuple[int, int]:
        """Return shape of the transformed DataFrame."""
        if is_arrow_dict(self.columns[0]):
            # When input is arrow.
            n_samples = len(self.columns[0].indices)
        elif _is_df_cat(self.columns[0]):
            # When input is pandas.
            n_samples = self.columns[0].codes.shape[0]
        else:
            # Anything else, TypeGuard is ignored by mypy 1.15.0 for some reason
            n_samples = self.columns[0].shape[0]  # type: ignore[union-attr]
        return n_samples, len(self.columns)


def _transform_pandas_df(
    data: "PdDataFrame",
    enable_categorical: bool,
    feature_names: Optional[FeatureNames] = None,
    feature_types: Optional[Union[FeatureTypes, Categories]] = None,
    meta: Optional[str] = None,
) -> Tuple[PandasTransformed, Optional[FeatureNames], Optional[FeatureTypes]]:
    if meta and len(data.columns) > 1 and meta not in _matrix_meta:
        raise ValueError(f"DataFrame for {meta} cannot have multiple columns")
    if data.columns.has_duplicates:
        duplicates = data.columns[data.columns.duplicated()].unique().tolist()
        raise ValueError(
            f"Duplicate column names are not supported. Duplicates found: {duplicates}"
        )

    feature_types, ref_categories = get_ref_categories(feature_types)
    feature_names, feature_types = pandas_feature_info(
        data, meta, feature_names, feature_types, enable_categorical
    )

    arrays = pandas_transform_data(data)
    return (
        PandasTransformed(arrays, ref_categories=ref_categories),
        feature_names,
        feature_types,
    )


def _meta_from_pandas_df(
    data: DataType,
    name: str,
    dtype: Optional[NumpyDType],
    handle: ctypes.c_void_p,
) -> None:
    data, _, _ = _transform_pandas_df(data, False, meta=name)
    if len(data.columns) == 1:
        array = data.columns[0]
    else:
        array = np.stack(data.columns).T

    array, dtype = _ensure_np_dtype(array, dtype)
    _meta_from_numpy(array, name, dtype, handle)


def _reject_pd_sparse_col_split(
    data: "PdDataFrame", data_split_mode: DataSplitMode
) -> None:
    """Sparse pandas columns are not supported with column-wise data split."""
    if data_split_mode != DataSplitMode.COL:
        return
    for _, dtype in zip(data.columns, data.dtypes):
        if is_pd_sparse_dtype(dtype):
            raise ValueError("Column split does not support pandas sparse array.")


def _from_pandas_df(
    *,
    data: "PdDataFrame",
    enable_categorical: bool,
    missing: FloatCompatible,
    nthread: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    _reject_pd_sparse_col_split(data, data_split_mode)
    df, feature_names, feature_types = _transform_pandas_df(
        data, enable_categorical, feature_names, feature_types
    )

    handle = ctypes.c_void_p()
    _check_call(
        _LIB.XGDMatrixCreateFromColumnar(
            df.array_interface(),
            make_jcargs(
                nthread=nthread, missing=missing, data_split_mode=data_split_mode
            ),
            ctypes.byref(handle),
        )
    )
    return handle, feature_names, feature_types


def _meta_from_pandas_series(
    data: DataType, name: str, dtype: Optional[NumpyDType], handle: ctypes.c_void_p
) -> None:
    """Help transform pandas series for meta data like labels"""
    if is_pd_sparse_dtype(data.dtype):
        data = data.values.to_dense().astype(np.float32)
    elif is_pa_ext_dtype(data.dtype):
        data = pandas_pa_type(data)
    else:
        data = data.to_numpy(np.float32, na_value=np.nan)

    if is_pd_sparse_dtype(getattr(data, "dtype", data)):
        data = data.to_dense()  # type: ignore[union-attr]
    assert len(data.shape) == 1 or data.shape[1] == 0 or data.shape[1] == 1
    _meta_from_numpy(data, name, dtype, handle)


def _from_pandas_series(
    *,
    data: DataType,
    missing: FloatCompatible,
    nthread: int,
    enable_categorical: bool,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
) -> DispatchedDataBackendReturnType:
    if (data.dtype.name not in _pandas_dtype_mapper) and not (
        is_pd_cat_dtype(data.dtype) and enable_categorical
    ):
        _invalid_dataframe_dtype(data)
    if enable_categorical and is_pd_cat_dtype(data.dtype):
        data = data.cat.codes
    return _from_numpy_array(
        data=data.values.reshape(data.shape[0], 1).astype("float"),
        missing=missing,
        nthread=nthread,
        feature_names=feature_names,
        feature_types=feature_types,
    )


class ArrowTransformed(TransformedDf):
    """A storage class for transformed arrow table."""

    def __init__(
        self,
        columns: List[Union["pa.NumericArray", "pa.DictionaryArray"]],
        ref_categories: Optional[Categories] = None,
    ) -> None:
        self.columns = columns

        if TYPE_CHECKING:
            import pyarrow as pa
        else:
            pa = import_pyarrow()

        aitfs: AifType = []
        temporary_buffers = []

        def push_series(col: Union["pa.NumericArray", "pa.DictionaryArray"]) -> None:
            if isinstance(col, pa.DictionaryArray):
                cats = col.dictionary
                codes = col.indices
                if not isinstance(cats, (pa.StringArray, pa.LargeStringArray)):
                    raise TypeError(
                        "Only string-based categorical index is supported for arrow."
                    )
                jnames, jcodes, buf = arrow_cat_inf(cats, codes)
                temporary_buffers.append(buf)
                aitfs.append((jnames, jcodes))
            else:
                jdata = _arrow_array_inf(col)
                aitfs.append(jdata)

        for col in self.columns:
            push_series(col)

        super().__init__(
            ref_categories=ref_categories,
            aitfs=aitfs,
            temporary_buffers=temporary_buffers,
        )

    @property
    def shape(self) -> Tuple[int, int]:
        """Return shape of the transformed DataFrame."""
        return len(self.columns[0]), len(self.columns)


def _transform_arrow_table(
    data: "pa.Table",
    enable_categorical: bool,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
) -> Tuple[ArrowTransformed, Optional[FeatureNames], Optional[FeatureTypes]]:
    if TYPE_CHECKING:
        import pyarrow as pa
    else:
        pa = import_pyarrow()

    t_names, t_types = _arrow_feature_info(data)
    feature_types, ref_categories = get_ref_categories(feature_types)

    if feature_names is None:
        feature_names = t_names
    if feature_types is None:
        feature_types = t_types

    columns = []
    for cname in feature_names:
        col0 = data.column(cname)
        col: Union["pa.NumericArray", "pa.DictionaryArray"] = col0.combine_chunks()
        if isinstance(col, pa.BooleanArray):
            col = col.cast(pa.int8())  # bit-compressed array, not supported.
        if is_arrow_dict(col) and not enable_categorical:
            # None because the function doesn't know how to get the type info from arrow
            # table.
            _invalid_dataframe_dtype(None)
        columns.append(col)

    df_t = ArrowTransformed(columns, ref_categories=ref_categories)
    return df_t, feature_names, feature_types


def _from_arrow_table(  # pylint: disable=too-many-positional-arguments
    data: DataType,
    enable_categorical: bool,
    missing: FloatCompatible,
    n_threads: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    df_t, feature_names, feature_types = _transform_arrow_table(
        data, enable_categorical, feature_names, feature_types
    )
    handle = ctypes.c_void_p()
    _check_call(
        _LIB.XGDMatrixCreateFromColumnar(
            df_t.array_interface(),
            make_jcargs(
                nthread=n_threads, missing=missing, data_split_mode=data_split_mode
            ),
            ctypes.byref(handle),
        )
    )
    return handle, feature_names, feature_types


@functools.cache
def _arrow_dtype() -> Dict[DataType, str]:
    import pyarrow as pa

    mapping = {
        pa.int8(): "int",
        pa.int16(): "int",
        pa.int32(): "int",
        pa.int64(): "int",
        pa.uint8(): "int",
        pa.uint16(): "int",
        pa.uint32(): "int",
        pa.uint64(): "int",
        pa.float16(): "float",
        pa.float32(): "float",
        pa.float64(): "float",
        pa.bool_(): "i",
    }

    return mapping


def _arrow_feature_info(data: DataType) -> Tuple[List[str], List]:
    if TYPE_CHECKING:
        import pyarrow as pa
    else:
        pa = import_pyarrow()

    table: "pa.Table" = data
    names = table.column_names

    def map_type(name: str) -> str:
        col = table.column(name)
        if isinstance(col.type, pa.DictionaryType):
            return CAT_T  # pylint: disable=unreachable

        return _arrow_dtype()[col.type]

    types = list(map(map_type, names))
    return names, types


def _meta_from_arrow_table(
    data: DataType,
    name: str,
    dtype: Optional[NumpyDType],
    handle: ctypes.c_void_p,
) -> None:
    table: "pa.Table" = data
    _meta_from_pandas_df(table.to_pandas(), name=name, dtype=dtype, handle=handle)


def _check_pyarrow_for_polars() -> None:
    if not is_pyarrow_available():
        raise ImportError("`pyarrow` is required for polars.")


def _reject_polars_categorical(data: DataType) -> None:
    pl = import_polars()

    for name, dtype in zip(data.columns, data.dtypes):
        if isinstance(dtype, pl.Categorical):
            raise ValueError(
                "XGBoost does not support `polars.Categorical` because its "
                "encoding can be sparse. Use `polars.Enum` instead. "
                f"Invalid column: {name}",
            )


def _transform_polars_df(
    data: DataType,
    enable_categorical: bool,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
) -> Tuple[ArrowTransformed, Optional[FeatureNames], Optional[FeatureTypes]]:
    if _is_polars_lazyframe(data):
        df = data.collect()
        warnings.warn(
            "Using the default parameters for the polars `LazyFrame.collect`. Consider"
            " passing a realized `DataFrame` or `Series` instead.",
            UserWarning,
        )
    else:
        df = data

    _check_pyarrow_for_polars()
    _reject_polars_categorical(df)
    table = df.to_arrow()
    return _transform_arrow_table(
        table, enable_categorical, feature_names, feature_types
    )


def _from_polars_df(  # pylint: disable=too-many-positional-arguments
    data: DataType,
    enable_categorical: bool,
    missing: FloatCompatible,
    n_threads: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    df_t, feature_names, feature_types = _transform_polars_df(
        data, enable_categorical, feature_names, feature_types
    )
    handle = ctypes.c_void_p()
    _check_call(
        _LIB.XGDMatrixCreateFromColumnar(
            df_t.array_interface(),
            make_jcargs(
                nthread=n_threads, missing=missing, data_split_mode=data_split_mode
            ),
            ctypes.byref(handle),
        )
    )
    return handle, feature_names, feature_types


@functools.cache
def _lazy_load_cudf_is_cat() -> Callable[[Any], bool]:
    try:
        from cudf import CategoricalDtype

        def is_categorical_dtype(dtype: Any) -> bool:
            return isinstance(dtype, CategoricalDtype)

    except ImportError:
        try:
            from cudf.api.types import (  # type: ignore[no-redef]
                is_categorical_dtype,
            )
        except ImportError:
            from cudf.utils.dtypes import (  # type: ignore[no-redef]
                is_categorical_dtype,
            )

    return is_categorical_dtype


@functools.cache
def _lazy_load_cudf_is_bool() -> Callable[[Any], bool]:
    from cudf.api.types import is_bool_dtype

    return is_bool_dtype


class CudfTransformed(TransformedDf):
    """A storage class for transformed cuDF dataframe."""

    def __init__(
        self,
        columns: List[Union["PdSeries", DfCatAccessor]],
        ref_categories: Optional[Categories],
    ) -> None:
        self.columns = columns
        # Buffers for temporary data that cannot be freed until the data is consumed by
        # the DMatrix or the booster.

        aitfs: AifType = []
        temporary_buffers = []

        def push_series(ser: Any) -> None:
            if _is_df_cat(ser):
                cats, codes = ser.categories, ser.codes
                cats_ainf, codes_ainf, buf = cudf_cat_inf(cats, codes)
                temporary_buffers.append(buf)
                aitfs.append((cats_ainf, codes_ainf))
            else:
                # numeric column
                ainf = cuda_array_interface_dict(ser)
                aitfs.append(ainf)

        for col in self.columns:
            push_series(col)

        super().__init__(
            ref_categories=ref_categories,
            aitfs=aitfs,
            temporary_buffers=temporary_buffers,
        )

    @property
    def shape(self) -> Tuple[int, int]:
        """Return shape of the transformed DataFrame."""
        if _is_df_cat(self.columns[0]):
            n_samples = self.columns[0].codes.shape[0]
        else:
            n_samples = self.columns[0].shape[0]  # type: ignore[union-attr]
        return n_samples, len(self.columns)


def _transform_cudf_df(
    data: DataType,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
    enable_categorical: bool,
) -> Tuple[
    CudfTransformed,
    Optional[FeatureNames],
    Optional[FeatureTypes],
]:
    is_bool_dtype = _lazy_load_cudf_is_bool()

    is_categorical_dtype = _lazy_load_cudf_is_cat()
    # Work around https://github.com/dmlc/xgboost/issues/10181
    if _is_cudf_ser(data):
        if is_bool_dtype(data.dtype):
            data = data.astype(np.uint8)
        dtypes = [data.dtype]
    else:
        data = data.astype(
            {col: np.uint8 for col in data.select_dtypes(include="bool")}
        )
        dtypes = data.dtypes

    if not all(
        dtype.name in _pandas_dtype_mapper
        or (is_categorical_dtype(dtype) and enable_categorical)
        for dtype in dtypes
    ):
        _invalid_dataframe_dtype(data)

    # handle feature names
    if feature_names is None:
        if _is_cudf_ser(data):
            feature_names = [data.name]
        elif lazy_isinstance(data.columns, "cudf.core.multiindex", "MultiIndex"):
            feature_names = [" ".join([str(x) for x in i]) for i in data.columns]
        else:
            feature_names = list(data.columns.map(str))

    # handle feature types
    feature_types, ref_categories = get_ref_categories(feature_types)
    if feature_types is None:
        feature_types = []
        for dtype in dtypes:
            if is_categorical_dtype(dtype) and enable_categorical:
                feature_types.append(CAT_T)
            else:
                feature_types.append(_pandas_dtype_mapper[dtype.name])

    # handle categorical data
    result = []
    if _is_cudf_ser(data):
        # unlike pandas, cuDF uses NA for missing data.
        if is_categorical_dtype(data.dtype) and enable_categorical:
            result.append(data.cat)
        elif is_categorical_dtype(data.dtype) and not enable_categorical:
            raise ValueError(_ENABLE_CAT_ERR)
        else:
            result.append(data)
    else:
        for col, dtype in zip(data.columns, data.dtypes):
            series = data[col]
            if is_categorical_dtype(dtype) and enable_categorical:
                result.append(series.cat)
            elif is_categorical_dtype(dtype):
                raise ValueError(_ENABLE_CAT_ERR)
            else:
                result.append(series)

    return (
        CudfTransformed(result, ref_categories=ref_categories),
        feature_names,
        feature_types,
    )


def _from_cudf_df(
    *,
    data: DataType,
    missing: FloatCompatible,
    nthread: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
    enable_categorical: bool,
) -> DispatchedDataBackendReturnType:
    df, feature_names, feature_types = _transform_cudf_df(
        data, feature_names, feature_types, enable_categorical
    )
    handle = ctypes.c_void_p()
    _check_call(
        _LIB.XGDMatrixCreateFromCudaColumnar(
            df.array_interface(),
            make_jcargs(nthread=nthread, missing=missing),
            ctypes.byref(handle),
        )
    )
    return handle, feature_names, feature_types


def _transform_cupy_array(data: DataType) -> CupyT:
    import cupy

    if not hasattr(data, "__cuda_array_interface__") and hasattr(data, "__array__"):
        data = cupy.array(data, copy=False)
    if array_hasobject(data) or data.dtype in [cupy.bool_]:
        data = data.astype(cupy.float32, copy=False)
    return data


def _from_cupy_array(
    data: DataType,
    missing: FloatCompatible,
    nthread: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
) -> DispatchedDataBackendReturnType:
    """Initialize DMatrix from cupy ndarray."""
    data = _transform_cupy_array(data)
    interface_str = cuda_array_interface(data)
    handle = ctypes.c_void_p()
    config = bytes(json.dumps({"missing": missing, "nthread": nthread}), "utf-8")
    _check_call(
        _LIB.XGDMatrixCreateFromCudaArrayInterface(
            interface_str, config, ctypes.byref(handle)
        )
    )
    return handle, feature_names, feature_types


def _is_cupy_csr(data: DataType) -> bool:
    try:
        import cupyx
    except ImportError:
        return False
    return isinstance(data, cupyx.scipy.sparse.csr_matrix)


def _is_cupy_csc(data: DataType) -> bool:
    try:
        import cupyx
    except ImportError:
        return False
    return isinstance(data, cupyx.scipy.sparse.csc_matrix)


def _is_dlpack(data: DataType) -> bool:
    return "PyCapsule" in str(type(data)) and "dltensor" in str(data)


def _transform_dlpack(data: DataType) -> bool:
    from cupy import from_dlpack  # pylint: disable=E0401

    assert "used_dltensor" not in str(data)
    data = from_dlpack(data)
    return data


def _from_dlpack(
    data: DataType,
    missing: FloatCompatible,
    nthread: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
) -> DispatchedDataBackendReturnType:
    data = _transform_dlpack(data)
    return _from_cupy_array(data, missing, nthread, feature_names, feature_types)


def _is_uri(data: DataType) -> TypeGuard[PathLike]:
    return isinstance(data, (str, os.PathLike))


def _from_uri(
    data: PathLike,
    missing: Optional[FloatCompatible],
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    _warn_unused_missing(data, missing)
    handle = ctypes.c_void_p()
    data = os.fspath(os.path.expanduser(data))
    config = make_jcargs(uri=str(data), data_split_mode=int(data_split_mode))
    _check_call(_LIB.XGDMatrixCreateFromURI(config, ctypes.byref(handle)))
    return handle, feature_names, feature_types


def _is_list(data: DataType) -> TypeGuard[list]:
    return isinstance(data, list)


def _from_list(
    *,
    data: Sequence,
    missing: FloatCompatible,
    n_threads: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    array = np.array(data)
    _check_data_shape(data)
    return _from_numpy_array(
        data=array,
        missing=missing,
        nthread=n_threads,
        feature_names=feature_names,
        feature_types=feature_types,
        data_split_mode=data_split_mode,
    )


def _is_tuple(data: DataType) -> TypeGuard[tuple]:
    return isinstance(data, tuple)


def _from_tuple(
    *,
    data: Sequence,
    missing: FloatCompatible,
    n_threads: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    return _from_list(
        data=data,
        missing=missing,
        n_threads=n_threads,
        feature_names=feature_names,
        feature_types=feature_types,
        data_split_mode=data_split_mode,
    )


def _has_array_protocol(data: DataType) -> bool:
    return hasattr(data, "__array__")


def _convert_unknown_data(data: DataType) -> DataType:
    warnings.warn(
        f"Unknown data type: {type(data)}, trying to convert it to csr_matrix",
        UserWarning,
    )
    try:
        import scipy.sparse
    except ImportError:
        return None

    try:
        data = scipy.sparse.csr_matrix(data)
    except Exception:  # pylint: disable=broad-except
        return None

    return data


def dispatch_data_backend(
    *,
    data: DataType,
    missing: FloatCompatible,  # Or Optional[Float]
    threads: int,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[Union[FeatureTypes, Categories]],
    enable_categorical: bool = False,
    data_split_mode: DataSplitMode = DataSplitMode.ROW,
) -> DispatchedDataBackendReturnType:
    """Dispatch data for DMatrix."""

    def check_cats(
        feature_types: Optional[Union[FeatureTypes, Categories]],
    ) -> TypeGuard[Optional[FeatureTypes]]:
        if isinstance(feature_types, Categories):
            raise ValueError(
                "Reference category is only supported by DataFrame inputs."
            )
        return True

    if (
        not _is_cudf_ser(data)
        and not _is_pandas_series(data)
        and not _is_polars_series(data)
    ):
        _check_data_shape(data)
    if is_scipy_csr(data):
        assert check_cats(feature_types)
        return _from_scipy_csr(
            data=data,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if is_scipy_csc(data):
        assert check_cats(feature_types)
        return _from_scipy_csc(
            data=data,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if is_scipy_coo(data):
        assert check_cats(feature_types)
        return _from_scipy_csr(
            data=data.tocsr(),
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if _is_np_array_like(data):
        assert check_cats(feature_types)
        return _from_numpy_array(
            data=data,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if _is_uri(data):
        assert check_cats(feature_types)
        return _from_uri(data, missing, feature_names, feature_types, data_split_mode)
    if _is_list(data):
        assert check_cats(feature_types)
        return _from_list(
            data=data,
            missing=missing,
            n_threads=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if _is_tuple(data):
        assert check_cats(feature_types)
        return _from_tuple(
            data=data,
            missing=missing,
            n_threads=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if _is_polars_series(data):
        pl = import_polars()

        data = pl.DataFrame({data.name: data})
    if _is_polars(data):
        return _from_polars_df(
            data,
            enable_categorical,
            missing=missing,
            n_threads=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if _is_arrow(data):
        return _from_arrow_table(
            data,
            enable_categorical,
            missing=missing,
            n_threads=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if _is_cudf_pandas(data):
        data = data._fsproxy_fast  # pylint: disable=protected-access
    if _is_pandas_series(data):
        pd = import_pandas()

        data = pd.DataFrame(data)
    if _is_pandas_df(data):
        return _from_pandas_df(
            data=data,
            enable_categorical=enable_categorical,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            data_split_mode=data_split_mode,
        )
    if _is_cudf_df(data) or _is_cudf_ser(data):
        return _from_cudf_df(
            data=data,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
            enable_categorical=enable_categorical,
        )
    if _is_cupy_alike(data):
        assert check_cats(feature_types)
        return _from_cupy_array(data, missing, threads, feature_names, feature_types)
    if _is_cupy_csr(data):
        raise TypeError("cupyx CSR is not supported yet.")
    if _is_cupy_csc(data):
        raise TypeError("cupyx CSC is not supported yet.")
    if _is_dlpack(data):
        assert check_cats(feature_types)
        return _from_dlpack(data, missing, threads, feature_names, feature_types)
    if _is_modin_series(data):
        pd = import_pandas()

        data = pd.DataFrame(data)
    if _is_modin_df(data):
        return _from_pandas_df(
            data=data,
            enable_categorical=enable_categorical,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
        )

    if _has_array_protocol(data):
        assert check_cats(feature_types)
        array = np.asarray(data)
        return _from_numpy_array(
            data=array,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
        )

    converted = _convert_unknown_data(data)
    if converted is not None:
        assert check_cats(feature_types)
        return _from_scipy_csr(
            data=converted,
            missing=missing,
            nthread=threads,
            feature_names=feature_names,
            feature_types=feature_types,
        )

    raise TypeError("Not supported type for data." + str(type(data)))


def _validate_meta_shape(data: DataType, name: str) -> None:
    if hasattr(data, "shape"):
        msg = f"Invalid shape: {data.shape} for {name}"
        if name in _matrix_meta:
            if len(data.shape) > 2:
                raise ValueError(msg)
            return

        if len(data.shape) > 2 or (
            len(data.shape) == 2 and (data.shape[1] != 0 and data.shape[1] != 1)
        ):
            raise ValueError(f"Invalid shape: {data.shape} for {name}")


def _meta_from_numpy(
    data: np.ndarray,
    field: str,
    dtype: Optional[NumpyDType],
    handle: ctypes.c_void_p,
) -> None:
    data, dtype = _ensure_np_dtype(data, dtype)
    interface = data.__array_interface__
    if interface.get("mask", None) is not None:
        raise ValueError("Masked array is not supported.")
    interface_str = array_interface(data)
    _check_call(_LIB.XGDMatrixSetInfoFromInterface(handle, c_str(field), interface_str))


def _meta_from_list(
    data: Sequence, field: str, dtype: Optional[NumpyDType], handle: ctypes.c_void_p
) -> None:
    data_np = np.array(data)
    _meta_from_numpy(data_np, field, dtype, handle)


def _meta_from_tuple(
    data: Sequence, field: str, dtype: Optional[NumpyDType], handle: ctypes.c_void_p
) -> None:
    return _meta_from_list(data, field, dtype, handle)


def _meta_from_cudf_df(data: DataType, field: str, handle: ctypes.c_void_p) -> None:
    if field not in _matrix_meta:
        _meta_from_cudf_series(data.iloc[:, 0], field, handle)
    else:
        data = data.values
        interface = cuda_array_interface(data)
        _check_call(_LIB.XGDMatrixSetInfoFromInterface(handle, c_str(field), interface))


def _meta_from_cudf_series(data: DataType, field: str, handle: ctypes.c_void_p) -> None:
    check_cudf_meta(data, field)
    inf = cuda_array_interface(data)
    _check_call(_LIB.XGDMatrixSetInfoFromInterface(handle, c_str(field), inf))


def _meta_from_cupy_array(data: DataType, field: str, handle: ctypes.c_void_p) -> None:
    data = _transform_cupy_array(data)
    inf = cuda_array_interface(data)
    _check_call(_LIB.XGDMatrixSetInfoFromInterface(handle, c_str(field), inf))


def dispatch_meta_backend(
    matrix: "DMatrix", data: DataType, name: str, dtype: Optional[NumpyDType] = None
) -> None:
    """Dispatch for meta info."""
    handle = matrix.handle
    assert handle is not None
    _validate_meta_shape(data, name)
    if data is None:
        return
    if _is_list(data):
        _meta_from_list(data, name, dtype, handle)
        return
    if _is_tuple(data):
        _meta_from_tuple(data, name, dtype, handle)
        return
    if _is_np_array_like(data):
        _meta_from_numpy(data, name, dtype, handle)
        return
    if _is_arrow(data):
        _meta_from_arrow_table(data, name, dtype, handle)
        return
    if _is_cudf_pandas(data):
        data = data._fsproxy_fast  # pylint: disable=protected-access
    if _is_polars(data):
        if _is_polars_lazyframe(data):
            data = data.collect()
        _check_pyarrow_for_polars()
        _meta_from_arrow_table(data.to_arrow(), name, dtype, handle)
        return
    if _is_pandas_df(data):
        _meta_from_pandas_df(data, name, dtype=dtype, handle=handle)
        return
    if _is_pandas_series(data):
        _meta_from_pandas_series(data, name, dtype, handle)
        return
    if _is_dlpack(data):
        data = _transform_dlpack(data)
        _meta_from_cupy_array(data, name, handle)
        return
    if _is_cudf_ser(data):
        _meta_from_cudf_series(data, name, handle)
        return
    if _is_cudf_df(data):
        _meta_from_cudf_df(data, name, handle)
        return
    if _is_cupy_alike(data):
        _meta_from_cupy_array(data, name, handle)
        return
    if _is_modin_df(data):
        _meta_from_pandas_df(data, name, dtype=dtype, handle=handle)
        return
    if _is_modin_series(data):
        data = data.values.astype("float")
        assert len(data.shape) == 1 or data.shape[1] == 0 or data.shape[1] == 1
        _meta_from_numpy(data, name, dtype, handle)
        return
    if _has_array_protocol(data):
        # pyarrow goes here.
        array = np.asarray(data)
        _meta_from_numpy(array, name, dtype, handle)
        return
    raise TypeError("Unsupported type for " + name, str(type(data)))


def _proxy_transform(
    data: DataType,
    feature_names: Optional[FeatureNames],
    feature_types: Optional[FeatureTypes],
    enable_categorical: bool,
) -> TransformedData:
    if _is_cudf_pandas(data):
        data = data._fsproxy_fast  # pylint: disable=protected-access
    if _is_cudf_df(data) or _is_cudf_ser(data):
        return _transform_cudf_df(
            data, feature_names, feature_types, enable_categorical
        )
    if _is_cupy_alike(data):
        data = _transform_cupy_array(data)
        return data, feature_names, feature_types
    if _is_dlpack(data):
        return _transform_dlpack(data), feature_names, feature_types
    if _is_list(data) or _is_tuple(data):
        data = np.array(data)
    if _is_np_array_like(data):
        data, _ = _ensure_np_dtype(data, data.dtype)
        return data, feature_names, feature_types
    if is_scipy_csr(data):
        data = transform_scipy_sparse(data, True)
        return data, feature_names, feature_types
    if is_scipy_csc(data):
        data = transform_scipy_sparse(data.tocsr(), True)
        return data, feature_names, feature_types
    if is_scipy_coo(data):
        data = transform_scipy_sparse(data.tocsr(), True)
        return data, feature_names, feature_types
    if _is_polars(data):
        df_pl, feature_names, feature_types = _transform_polars_df(
            data, enable_categorical, feature_names, feature_types
        )
        return df_pl, feature_names, feature_types
    if _is_pandas_series(data):
        pd = import_pandas()

        data = pd.DataFrame(data)
    if _is_arrow(data):
        df_pa, feature_names, feature_types = _transform_arrow_table(
            data, enable_categorical, feature_names, feature_types
        )
        return df_pa, feature_names, feature_types
    if _is_pandas_df(data):
        df, feature_names, feature_types = _transform_pandas_df(
            data, enable_categorical, feature_names, feature_types
        )
        return df, feature_names, feature_types
    raise TypeError("Value type is not supported for data iterator:" + str(type(data)))


def is_on_cuda(data: Any) -> bool:
    """Whether the data is a CUDA-based data structure."""
    return any(
        p(data)
        for p in (
            _is_cudf_df,
            _is_cudf_ser,
            _is_cudf_pandas,
            _is_cupy_alike,
            _is_dlpack,
        )
    )


def dispatch_proxy_set_data(
    proxy: "_ProxyDMatrix",
    data: DataType,
) -> None:
    """Dispatch for QuantileDMatrix."""
    if (
        not _is_cudf_ser(data)
        and not _is_pandas_series(data)
        and not _is_polars_series(data)
    ):
        _check_data_shape(data)

    if isinstance(data, CudfTransformed):
        # pylint: disable=W0212
        proxy._ref_data_from_cuda_columnar(data)
        return
    if _is_cupy_alike(data):
        proxy._ref_data_from_cuda_interface(data)  # pylint: disable=W0212
        return
    if _is_dlpack(data):
        data = _transform_dlpack(data)
        proxy._ref_data_from_cuda_interface(data)  # pylint: disable=W0212
        return
    # Host
    if isinstance(data, (ArrowTransformed, PandasTransformed)):
        proxy._ref_data_from_columnar(data)  # pylint: disable=W0212
        return
    if _is_np_array_like(data):
        _check_data_shape(data)
        proxy._ref_data_from_array(data)  # pylint: disable=W0212
        return
    if is_scipy_csr(data):
        proxy._ref_data_from_csr(data)  # pylint: disable=W0212
        return

    err = TypeError("Value type is not supported for data iterator:" + str(type(data)))
    raise err



"""Tests for the CUDA implementation of multi-target."""

# pylint: disable=too-many-positional-arguments,missing-function-docstring
from typing import Any, Callable, Dict, Optional

import numpy as np
import pytest
import xgboost as xgb
from hypothesis import given, note, settings, strategies
from xgboost import config_context
from xgboost import testing as tm
from xgboost.testing.multi_target import (
    all_reg_objectives,
    run_absolute_error,
    run_column_sampling,
    run_deterministic,
    run_eta,
    run_expectile_loss,
    run_feature_importance_strategy_compare,
    run_gradient_based_sampling_accuracy,
    run_grow_policy,
    run_mixed_strategy,
    run_multiclass,
    run_multilabel,
    run_quantile_loss,
    run_reduced_grad,
    run_subsample,
    run_with_iter,
)
from xgboost.testing.params import hist_parameter_strategy
from xgboost.testing.updater import check_quantile_loss_rf, train_result
from xgboost.testing.utils import Device


@pytest.mark.parametrize("learning_rate", [1.0, None])
def test_multiclass(learning_rate: Optional[float]) -> None:
    run_multiclass("cuda", learning_rate)


@pytest.mark.parametrize("learning_rate", [1.0, None])
def test_multilabel(learning_rate: Optional[float]) -> None:
    run_multilabel("cuda", learning_rate)


@pytest.mark.parametrize("weighted", [True, False])
def test_quantile_loss(weighted: bool) -> None:
    run_quantile_loss("cuda", weighted)


def test_expectile_loss() -> None:
    run_expectile_loss("cuda")


@pytest.mark.parametrize("multi_strategy", ["multi_output_tree", "one_output_per_tree"])
def test_quantile_loss_rf(multi_strategy: str) -> None:
    check_quantile_loss_rf("cuda", "hist", multi_strategy)
    if multi_strategy == "one_output_per_tree":
        check_quantile_loss_rf("cuda", "approx", multi_strategy)


def test_shap_multi_output_tree() -> None:
    X = np.array(
        [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0], [2.0, 1.0], [1.0, 2.0]],
        dtype=np.float32,
    )
    y = np.stack([X[:, 0] + 2 * X[:, 1], -X[:, 0] + 0.5 * X[:, 1]], axis=1)
    Xy = xgb.DMatrix(X, y)

    booster = xgb.train(
        {
            "device": "cuda",
            "tree_method": "hist",
            "max_depth": 2,
            "num_target": 2,
            "multi_strategy": "multi_output_tree",
            "objective": "reg:squarederror",
        },
        Xy,
        num_boost_round=3,
    )

    margin = booster.predict(Xy, output_margin=True, strict_shape=True)
    contribs = booster.predict(Xy, pred_contribs=True, strict_shape=True)
    interactions = booster.predict(Xy, pred_interactions=True, strict_shape=True)

    assert margin.shape == (X.shape[0], y.shape[1])
    assert contribs.shape == (X.shape[0], y.shape[1], X.shape[1] + 1)
    assert interactions.shape == (
        X.shape[0],
        y.shape[1],
        X.shape[1] + 1,
        X.shape[1] + 1,
    )
    np.testing.assert_allclose(contribs.sum(axis=2), margin, rtol=1e-4, atol=1e-4)
    np.testing.assert_allclose(
        interactions.sum(axis=(2, 3)), margin, rtol=1e-4, atol=1e-4
    )


def test_absolute_error() -> None:
    run_absolute_error("cuda")


def test_reduced_grad() -> None:
    run_reduced_grad("cuda")


def test_with_iter() -> None:
    with config_context(use_rmm=True):
        run_with_iter("cuda")


def test_eta() -> None:
    run_eta("cuda")


def test_deterministic() -> None:
    run_deterministic("cuda")


def test_column_sampling() -> None:
    run_column_sampling("cuda")


@pytest.mark.parametrize("grow_policy", ["depthwise", "lossguide"])
def test_grow_policy(grow_policy: str) -> None:
    run_grow_policy("cuda", grow_policy)


def test_mixed_strategy() -> None:
    run_mixed_strategy("cuda")


def test_feature_importance_strategy_compare() -> None:
    run_feature_importance_strategy_compare("cuda")


@given(hist_parameter_strategy, strategies.integers(1, 20), tm.multi_dataset_strategy)
@settings(deadline=None, max_examples=50, print_blob=True)
def test_hist(param: Dict[str, Any], num_rounds: int, dataset: tm.TestDataset) -> None:
    param["tree_method"] = "hist"
    param["device"] = "cuda"
    param = dataset.set_params(param)
    result = train_result(param, dataset.get_dmat(), num_rounds)
    note(str(result))
    assert tm.non_increasing(result["train"][dataset.metric])


@pytest.mark.parametrize("obj_fn", all_reg_objectives())
def test_reg_objective(obj_fn: Callable[[Device], None]) -> None:
    obj_fn("cuda")


@pytest.mark.parametrize("sampling_method", ["uniform", "gradient_based"])
def test_subsample(sampling_method: str) -> None:
    run_subsample("cuda", sampling_method)


def test_gradient_based_sampling_accuracy() -> None:
    run_gradient_based_sampling_accuracy("cuda")
